# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen

import warnings

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.exceptions import NotFittedError

from greykite.common.features.timeseries_impute import impute_with_lags_multi
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.python_utils import update_dictionary


DEFAULT_PARAMS = {
    "interpolate": dict(
        method="linear",
        limit_direction="both",
        axis=0),  # fills column-by-column
    "ts_interpolate": dict(
        orders=[7, 14, 21],
        agg_func=np.mean,
        iter_num=5)
}


class NullTransformer(BaseEstimator, TransformerMixin):
    """Imputes nulls in time series data.

    This transform is stateless in the sense that ``transform`` output
    does not depend on the data passed to ``fit``. The dataset passed to
    ``transform`` is used to impute itself.

    Parameters
    ----------
    max_frac : `float`, default 0.10
        issues warning if fraction of nulls is above this value
    impute_algorithm  : `str` or None, default "interpolate"
        Which imputation algorithm to use.
        Valid options are:

            - "interpolate" : `pandas.DataFrame.interpolate`
            - "ts_interpolate" : `~greykite.common.features.timeseries_impute.impute_with_lags_multi`.

        If None, this transformer is a no-op. No null imputation is done.

    impute_params : `dict` or None, default None
        Params to pass to the imputation algorithm.
        See `pandas.DataFrame.interpolate` and
        `~greykite.common.features.timeseries_impute.impute_with_lags_multi`
        for their respective options.

        For pandas "interpolate", the "ffill", "pad", "bfill", "backfill" methods
        are not allowed to avoid confusion with the fill axis parameter. Use "linear"
        with ``axis=0`` instead, with direction controlled by ``limit_direction``.

        If None, uses the defaults provided in this class.
    impute_all : `bool`, default True
        Whether to impute all values. If True, NaNs are not allowed in the
        transformed result. Ignored if ``impute_algorithm`` is None.

        The transform specified by ``impute_algorithm`` and
        ``impute_params`` may leave NaNs in the dataset. For example,
        if it fills in the forward direction but the first value in a
        column is NaN.

        A first pass is taken with the impute algorithm specified.
        A second pass is taken with the "interpolate" algorithm (method="linear",
        limit_direction="both") to fill in remaining NaNs.

    Attributes
    ----------
    null_frac : `int`
        The fraction data points that are null
    _is_fitted : `bool`
        Whether the transformer is fitted.
    missing_info : `dict`
        Information about the missing data.
        Set by ``transform`` if ``impute_algorithm = "ts_interpolate"``.
    """
    def __init__(
            self,
            max_frac=0.10,
            impute_algorithm=None,
            impute_params=None,
            impute_all=True):
        # sets params without modification to ensure get_params() works in grid search
        self.max_frac = max_frac
        self.impute_algorithm = impute_algorithm
        self.impute_params = impute_params
        self.impute_all = impute_all

        self.null_frac = None
        self._is_fitted = None
        self.missing_info = None
        if (self.impute_algorithm == "interpolate"
                and self.impute_params is not None
                and self.impute_params.get("method") in ["ffill", "pad", "bfill", "backfill"]):
            # These four methods treat "axis=0" as rows, "axis=1" as columns,
            # contrary to the pandas documentation. Avoid them to prevent misuse.
            raise ValueError(
                f"method '{self.impute_params['method']}' is not allowed. "
                f"Use method='linear' with `limit_direction` instead")

    def fit(self, X, y=None):
        """Updates `self.impute_params`.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Training input data. e.g. each column is a timeseries.
            Columns are expected to be numeric.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        assert isinstance(X, pd.DataFrame)
        self._is_fitted = True
        # sets default parameters
        if self.impute_algorithm is not None:
            default_params = DEFAULT_PARAMS.get(self.impute_algorithm, {})
            self.impute_params = update_dictionary(default_params, overwrite_dict=self.impute_params)
        return self

    def transform(self, X):
        """Imputes missing values in input time series.

        Checks the % of data points that are null, and provides warning if
        it exceeds ``self.max_frac``.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Data to transform. e.g. each column is a timeseries.
            Columns are expected to be numeric.

        Returns
        -------
        X_imputed : `pandas.DataFrame`
            A copy of the data frame with original values and missing values imputed
        """
        if self._is_fitted is None:
            raise NotFittedError(
                "This instance is not fitted yet. Call 'fit' with appropriate arguments "
                "before calling 'transform'.")
        assert isinstance(X, pd.DataFrame)

        self.null_frac = X.isna().mean()  # fraction of NaNs in each column
        if np.any(self.null_frac > self.max_frac):
            warnings.warn(f"Input data has many null values. Missing {self.null_frac.max():.2%} of one input.",
                          RuntimeWarning)
        if any(self.null_frac > 0.0):
            log_message(f"Missing data detected: {self.null_frac.mean():.2%} of all input values "
                        f"are null. (If future external regressor(s) are used, some missing values in "
                        f"`value_col` are expected.)",
                        LoggingLevelEnum.INFO)

        if self.impute_algorithm is not None:
            if self.impute_algorithm == "interpolate":
                # Uses `pandas.DataFrame.interpolate`
                X_imputed = X.interpolate(**self.impute_params)
            elif self.impute_algorithm == "ts_interpolate":
                # Uses `impute_with_lags_multi`
                impute_info = impute_with_lags_multi(df=X, **self.impute_params)
                X_imputed = impute_info["df"]
                self.missing_info = impute_info["missing_info"]
            else:
                raise ValueError(f"`impute_algorithm` '{self.impute_algorithm}' is not recognized."
                                 f"Must be one of 'ts_interpolate', 'interpolate'")

            if self.impute_all:
                # A second pass is taken to make sure there are no NaNs.
                X_imputed = X_imputed.interpolate(**DEFAULT_PARAMS["interpolate"])
        else:
            # no-op
            X_imputed = X.copy()
        return X_imputed



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""sklearn estimator for ``simple_forecast_silverkite``"""

from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from sklearn.metrics import mean_squared_error

from greykite.algo.forecast.silverkite.forecast_simple_silverkite import SimpleSilverkiteForecast
from greykite.algo.forecast.silverkite.silverkite_diagnostics import SilverkiteDiagnostics
from greykite.common import constants as cst
from greykite.common.python_utils import update_dictionary
from greykite.sklearn.estimator.base_silverkite_estimator import BaseSilverkiteEstimator


class SimpleSilverkiteEstimator(BaseSilverkiteEstimator):
    """Wrapper for `~greykite.algo.forecast.silverkite.forecast_simple_silverkite.forecast_simple_silverkite`.

    Parameters
    ----------
    score_func : callable, optional, default mean_squared_error
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    coverage : `float` between [0.0, 1.0] or None, optional
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    null_model_params : `dict` or None, optional
        Dictionary with arguments to define ``DummyRegressor`` null model, default is `None`.
        See `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
    fit_algorithm_dict : `dict` or None, optional
        How to fit the model. A dictionary with the following optional keys.

            ``"fit_algorithm"`` : `str`, optional, default "ridge"
                The type of predictive model used in fitting.

                See `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
                for available options and their parameters.
            ``"fit_algorithm_params"`` : `dict` or None, optional, default None
                Parameters passed to the requested fit_algorithm.
                If None, uses the defaults in `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.

    uncertainty_dict : `dict` or `str` or None, optional
        How to fit the uncertainty model.
        See `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
        Note that this is allowed to be "auto". If None or "auto", will be set to
        a default value by ``coverage`` before calling ``forecast_silverkite``.
        See ``BaseForecastEstimator`` for details.

    kwargs : additional parameters

        Other parameters are the same as in
        `~greykite.algo.forecast.silverkite.forecast_simple_silverkite.forecast_simple_silverkite`.

        See source code ``__init__`` for the parameter names, and refer to
        `~greykite.algo.forecast.silverkite.forecast_simple_silverkite.forecast_simple_silverkite` for
        their description.

        If this Estimator is called from
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`,
        ``train_test_thresh`` and ``training_fraction`` should almost
        always be `None`, because train/test is handled outside
        this Estimator.

    Notes
    -----
    Attributes match those of
    `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator`.

    See Also
    --------
    `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator`
        For attributes and details on fit, predict, and component plots.
    `~greykite.algo.forecast.silverkite.forecast_simple_silverkite.forecast_simple_silverkite`
        Function to transform the parameters to call ``forecast_silverkite`` fit.
    `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`
        Functions performing the fit and predict.
    """
    def __init__(
            self,
            silverkite: SimpleSilverkiteForecast = SimpleSilverkiteForecast(),
            silverkite_diagnostics: SilverkiteDiagnostics = SilverkiteDiagnostics(),
            score_func: callable = mean_squared_error,
            coverage: float = None,
            null_model_params: Optional[Dict] = None,
            time_properties: Optional[Dict] = None,
            freq: Optional[str] = None,
            forecast_horizon: Optional[int] = None,
            origin_for_time_vars: Optional[float] = None,
            train_test_thresh: Optional[datetime] = None,
            training_fraction: Optional[float] = None,
            fit_algorithm_dict: Optional[Dict] = None,
            holidays_to_model_separately: Optional[Union[str, List[str]]] = "auto",
            holiday_lookup_countries: Optional[Union[str, List[str]]] = "auto",
            holiday_pre_num_days: int = 2,
            holiday_post_num_days: int = 2,
            holiday_pre_post_num_dict: Optional[Dict] = None,
            daily_event_df_dict: Optional[Dict] = None,
            changepoints_dict: Optional[Dict] = None,
            yearly_seasonality: Union[bool, str, int] = "auto",
            quarterly_seasonality: Union[bool, str, int] = "auto",
            monthly_seasonality: Union[bool, str, int] = "auto",
            weekly_seasonality: Union[bool, str, int] = "auto",
            daily_seasonality: Union[bool, str, int] = "auto",
            max_daily_seas_interaction_order: Optional[int] = None,
            max_weekly_seas_interaction_order: Optional[int] = None,
            autoreg_dict: Optional[Dict] = None,
            seasonality_changepoints_dict: Optional[Dict] = None,
            min_admissible_value: Optional[float] = None,
            max_admissible_value: Optional[float] = None,
            uncertainty_dict: Optional[Dict] = None,
            growth_term: Optional[str] = "linear",
            regressor_cols: Optional[List[str]] = None,
            feature_sets_enabled: Optional[Union[bool, Dict[str, bool]]] = None,
            extra_pred_cols: Optional[List[str]] = None,
            regression_weight_col: Optional[str] = None,
            simulation_based: Optional[bool] = False):
        # every subclass of BaseSilverkiteEstimator must call super().__init__
        super().__init__(
            silverkite=silverkite,
            silverkite_diagnostics=silverkite_diagnostics,
            score_func=score_func,
            coverage=coverage,
            null_model_params=null_model_params,
            uncertainty_dict=uncertainty_dict)

        # necessary to set parameters, to ensure get_params() works
        # (used in grid search)
        self.score_func = score_func
        self.coverage = coverage
        self.null_model_params = null_model_params
        self.time_properties = time_properties
        self.freq = freq
        self.forecast_horizon = forecast_horizon
        self.origin_for_time_vars = origin_for_time_vars
        self.train_test_thresh = train_test_thresh
        self.training_fraction = training_fraction
        self.fit_algorithm_dict = fit_algorithm_dict
        self.holidays_to_model_separately = holidays_to_model_separately
        self.holiday_lookup_countries = holiday_lookup_countries
        self.holiday_pre_num_days = holiday_pre_num_days
        self.holiday_post_num_days = holiday_post_num_days
        self.holiday_pre_post_num_dict = holiday_pre_post_num_dict
        self.daily_event_df_dict = daily_event_df_dict
        self.changepoints_dict = changepoints_dict
        self.yearly_seasonality = yearly_seasonality
        self.quarterly_seasonality = quarterly_seasonality
        self.monthly_seasonality = monthly_seasonality
        self.weekly_seasonality = weekly_seasonality
        self.daily_seasonality = daily_seasonality
        self.max_daily_seas_interaction_order = max_daily_seas_interaction_order
        self.max_weekly_seas_interaction_order = max_weekly_seas_interaction_order
        self.autoreg_dict = autoreg_dict
        self.seasonality_changepoints_dict = seasonality_changepoints_dict
        self.min_admissible_value = min_admissible_value
        self.max_admissible_value = max_admissible_value
        self.uncertainty_dict = uncertainty_dict
        self.growth_term = growth_term
        self.regressor_cols = regressor_cols
        self.feature_sets_enabled = feature_sets_enabled
        self.extra_pred_cols = extra_pred_cols
        self.regression_weight_col = regression_weight_col
        self.simulation_based = simulation_based
        # ``forecast_simple_silverkite`` generates a ``fs_components_df`` to call
        # ``forecast_silverkite`` that is compatible with ``BaseSilverkiteEstimator``.
        # Unlike ``SilverkiteEstimator``, this does not need to call ``validate_inputs``.

    def fit(
            self,
            X,
            y=None,
            time_col=cst.TIME_COL,
            value_col=cst.VALUE_COL,
            **fit_params):
        """Fits ``Silverkite`` forecast model.

        Parameters
        ----------
        X: `pandas.DataFrame`
            Input timeseries, with timestamp column,
            value column, and any additional regressors.
            The value column is the response, included in
            ``X`` to allow transformation by `sklearn.pipeline`.
        y: ignored
            The original timeseries values, ignored.
            (The ``y`` for fitting is included in ``X``).
        time_col: `str`
            Time column name in ``X``.
        value_col: `str`
            Value column name in ``X``.
        fit_params: `dict`
            additional parameters for null model.

        Returns
        -------
        self : self
            Fitted model is stored in ``self.model_dict``.
        """
        # Initializes `fit_algorithm_dict` with default values.
        # This cannot be done in __init__ to remain compatible
        # with sklearn grid search.
        default_fit_algorithm_dict = {
            "fit_algorithm": "ridge",
            "fit_algorithm_params": None}
        self.fit_algorithm_dict = update_dictionary(
            default_fit_algorithm_dict,
            overwrite_dict=self.fit_algorithm_dict)

        # Fits null model
        super().fit(
            X=X,
            y=y,
            time_col=time_col,
            value_col=value_col,
            **fit_params)

        self.model_dict = self.silverkite.forecast_simple(
            df=X,
            time_col=time_col,
            value_col=value_col,
            time_properties=self.time_properties,
            freq=self.freq,
            forecast_horizon=self.forecast_horizon,
            origin_for_time_vars=self.origin_for_time_vars,
            train_test_thresh=self.train_test_thresh,
            training_fraction=self.training_fraction,
            fit_algorithm=self.fit_algorithm_dict["fit_algorithm"],
            fit_algorithm_params=self.fit_algorithm_dict["fit_algorithm_params"],
            holidays_to_model_separately=self.holidays_to_model_separately,
            holiday_lookup_countries=self.holiday_lookup_countries,
            holiday_pre_num_days=self.holiday_pre_num_days,
            holiday_post_num_days=self.holiday_post_num_days,
            holiday_pre_post_num_dict=self.holiday_pre_post_num_dict,
            daily_event_df_dict=self.daily_event_df_dict,
            changepoints_dict=self.changepoints_dict,
            yearly_seasonality=self.yearly_seasonality,
            quarterly_seasonality=self.quarterly_seasonality,
            monthly_seasonality=self.monthly_seasonality,
            weekly_seasonality=self.weekly_seasonality,
            daily_seasonality=self.daily_seasonality,
            max_daily_seas_interaction_order=self.max_daily_seas_interaction_order,
            max_weekly_seas_interaction_order=self.max_weekly_seas_interaction_order,
            autoreg_dict=self.autoreg_dict,
            seasonality_changepoints_dict=self.seasonality_changepoints_dict,
            min_admissible_value=self.min_admissible_value,
            max_admissible_value=self.max_admissible_value,
            uncertainty_dict=self.uncertainty_dict,
            growth_term=self.growth_term,
            regressor_cols=self.regressor_cols,
            feature_sets_enabled=self.feature_sets_enabled,
            extra_pred_cols=self.extra_pred_cols,
            regression_weight_col=self.regression_weight_col,
            simulation_based=self.simulation_based)

        # Sets attributes based on ``self.model_dict``
        super().finish_fit()
        return self



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Sliding window benchmarking."""

import warnings

import numpy as np
import pandas as pd


def gen_moving_timeseries_forecast(
        df,
        time_col,
        value_col,
        train_forecast_func,
        train_move_ahead,
        forecast_horizon,
        min_training_end_point=None,
        min_training_end_timestamp=None,
        max_forecast_end_point=None,
        max_forecast_end_timestamp=None,
        regressor_cols=None,
        keep_cols=None,  # extra cols in df which we want to keep
        **model_params):
    """Applies a forecast function (`train_forecast_func`) to many derived
    timeseries from `df` which are moving windows of `df`. For each derived
    series a model is trained and forecast is generated.
    It returns a `compare_df` to compare actuals and forecasts.

    Parameters
    ----------

    df : `pandas.DataFrame`
        A data frame which includes the timestamp column
        as well as the value column.
        The time column is assumed to be in increasing order (
        timestamps increase).
    time_col : `str`
        The column name in ``df`` representing time for the time series data.
        The time column can be anything that can be parsed by pandas DatetimeIndex.
    value_col: `str`
        The column name which has the value of interest to be forecasted.
    train_forecast_func : `func`
        A function with this signature::

        train_forecast_func(
            df,
            time_col,
            value_col,
            forecast_horizon,
            new_external_regressor_df=None)

        This function is required to return a dictionary which has at minimum this item
            "fut_df": pd.DataFrame which includes the forecasts in the column
            ``value_col``

    train_move_ahead : `int`
        The number of steps moving forward for each window
        This can be set to 1 often to get the maximum number of validations
        However other numbers can be used e.g. if computation is an issue
    forecast_horizon : `int`
        The number of forecasts needed
    min_training_end_point : `int` or None, default None
        The minimum number of training time points
    min_training_end_timestamp : `str` or None, default None
        The minimum timestamp to be used.
        If this is not None, ``min_training_end_point`` will be overwritten.
    max_forecast_end_point : `int` or None, default None
        The end point to be forecasted. The input ``df`` will be limited
        to this point.
    max_forecast_end_timestamp : `str` or None, default None
        The last timestamp allowed to be forecasted.
        If this is not None, ``max_forecast_end_point`` will be overwritten.
    regressor_cols : `list` [`str`] or None, default None
        If regressors are to be used, they are listed here.
    keep_cols : `list` [`str`] or None, default None
        Extra columns in ``df`` which we want to keep

    Return : `dict`
    ----------
    A dictionary with following items:

    - "compare_df": `pd.DataFrame`
        A dataframe which includes
        (a) actual true values (observed) given in "y_true" column;
        (b) forecasted values given in "y_hat";
        (c) horizon given in a column "horizon" which determines the number of
            points into the future for that forecast;
        (d) training end point given in ``training_end_point`` column
    - "max_possible_validation_num" : `int`
        Maximum possible number of validations
    - "validation_num" : `int`
        Number of validations used

    """
    if max_forecast_end_timestamp is not None:
        max_forecast_end_point = max(
            np.where(df[time_col] <= max_forecast_end_timestamp)[0])

    if min_training_end_timestamp is not None:
        min_training_end_point = min(
            np.where(df[time_col] >= min_training_end_timestamp)[0])

    if max_forecast_end_point is not None:
        df = df[:max_forecast_end_point]

    compare_df = None
    n = df.shape[0]
    if (n - forecast_horizon) <= min_training_end_point:
        raise ValueError("No reasonble train test period is found for validation")

    # Maximum possible validation number for this set
    max_possible_validation_num = n - forecast_horizon - min_training_end_point
    # Actual validation number
    validation_num = max_possible_validation_num / train_move_ahead

    training_end_times = np.arange(
        min_training_end_point,
        n - forecast_horizon,
        train_move_ahead)

    def get_compare_df_row(m):
        """Calculates comparison df with actuals and forecasted for the given
        horizon, using the training data up to time ``m``

        ----------
        Parameters
        m : `int`
            Last row of data to be used for training

        -------
        Returns
        compare_df0 : `pandas.DataFrame`
            A pandas dataframe with ``forecast_horizon`` rows containing
            observed values and forecasted values.
            It includes
                (a) actual true values (observed) given in "y_true" column;
                (b) forecasted values given in "y_hat";
                (c) horizon given in a column "horizon" which determines the number of
                    points into the future for that forecast;
                (d) training end point given in ``training_end_point`` column
        """
        train_df = df[:m]
        test_df = df.loc[range(m, m + forecast_horizon), :].reset_index(drop=True)

        if regressor_cols is not None:
            new_external_regressor_df = test_df[regressor_cols]
            obtained_forecast = train_forecast_func(
                df=train_df,
                value_col=value_col,
                time_col=time_col,
                forecast_horizon=forecast_horizon,
                new_external_regressor_df=new_external_regressor_df,
                **model_params)
        else:
            obtained_forecast = train_forecast_func(
                df=train_df,
                value_col=value_col,
                time_col=time_col,
                forecast_horizon=forecast_horizon,
                **model_params)

        y_hat = obtained_forecast["fut_df"][value_col].values
        y_true = test_df[value_col]

        timestamps = test_df[time_col]
        compare_df0 = pd.DataFrame({
            time_col: timestamps,
            "y_hat": y_hat,
            "y_true": y_true})

        if keep_cols is not None:
            compare_df0 = pd.concat(
                [compare_df0, test_df[keep_cols]],
                axis=1)
        compare_df0["horizon"] = range(1, forecast_horizon + 1)
        compare_df0["training_end_point"] = m

        return compare_df0
    # Runs the function for all m and stores the results dataframes for each m
    compare_df_list = [get_compare_df_row(m) for m in training_end_times]
    # Concats all the dataframes
    compare_df = pd.concat(compare_df_list, axis=0)

    na_df = compare_df[compare_df.isnull().any(axis=1)]
    if na_df.shape[0] > 0:
        warnings.warn("NA was generated in compare_df.")

    return {
        "compare_df": compare_df,
        "max_possible_validation_num": max_possible_validation_num,
        "validation_num": validation_num}



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""sklearn estimator for a dummy model that returns a constant forecast."""

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin
from sklearn.dummy import DummyRegressor
from sklearn.metrics import mean_squared_error

from greykite.common.constants import PREDICTED_COL
from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message


class DummyEstimator(BaseEstimator, RegressorMixin):
    """Pandas Wrapper for DummyRegressor. Unlike DummyRegressor, it uses X for fitting instead of y
    Otherwise, the interface is identical.

    This allows the DummyRegressor to be part of a Pipeline which transforms X[value_col] before fitting.

    Does not implement BaseForecastEstimator, because it is imported by that class.

    Parameters
    ----------
    strategy : str
        Strategy to use to generate predictions.

        * "mean": always predicts the mean of the training set
        * "median": always predicts the median of the training set
        * "quantile": always predicts a specified quantile of the training set,
          provided with the quantile parameter.
        * "constant": always predicts a constant value that is provided by
          the user.

    constant : int or float or array of shape = [n_outputs]
        The explicit constant as predicted by the "constant" strategy. This
        parameter is useful only for the "constant" strategy.

    quantile : float in [0.0, 1.0]
        The quantile to predict using the "quantile" strategy. A quantile of
        0.5 corresponds to the median, while 0.0 to the minimum and 1.0 to the
        maximum.

    score_func : callable, optional, default=mean_squared_error
        Function to calculate model R2_null_model_score score,
        with signature (actual, predicted).
        `actual`, `predicted` are array-like with the same shape.
        Smaller values are better.

    Attributes
    ----------
    coverage : None
        Model coverage.
        Always None, provided for compatibility with
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
    null_model: None
        Null model.
        Always None, provided for compatibility with
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
    model : `~sklearn.dummy.DummyRegressor`
        Trained model
    time_col_ : `str`
        Name of timestamp column
    value_col_ : `str`
        Name of value column
    """
    def __init__(self, strategy="mean", constant=None, quantile=None, score_func=mean_squared_error):
        self.strategy = strategy
        self.constant = constant
        self.quantile = quantile
        self.score_func = score_func

        # Always None, provided for compatibility with `forecast_pipeline`
        self.coverage = None
        self.null_model = None

        # Initializes attributes defined in fit
        self.model = None
        self.time_col_ = None
        self.value_col_ = None

    def fit(self, X, y=None, time_col=TIME_COL, value_col=VALUE_COL, sample_weight=None):
        """Fits the regressor using X[value_col]

        Parameters
        ----------
        X : pd.DataFrame with columns time_col and value_col
            Training data, requires length = n_samples

        y : ignored

        time_col: string, optional
            name of timestamp column in X

        value_col: string, optional
            name of value column in X

        sample_weight : array-like of shape = [n_samples], optional
            Sample weights.

        Returns
        -------
        self : object
        """
        self.time_col_ = time_col  # to be used in `predict` to select proper column
        self.value_col_ = value_col
        # Model cannot be initialized in __init__, otherwise scikit-learn
        # grid search will not be able to set the parameters.
        # See https://scikit-learn.org/stable/developers/develop.html#instantiation.
        self.model = DummyRegressor(
            strategy=self.strategy,
            constant=self.constant,
            quantile=self.quantile)

        # Drops records with missing values
        X = X.dropna(subset=[value_col])

        # Converts pd.DataFrame to np.array. Uses X[value_col] as y for fitting
        x_array = X[[value_col]].to_numpy()
        y_array = X[value_col].to_numpy()
        self.model.fit(X=x_array, y=y_array, sample_weight=sample_weight)
        return self

    def predict(self, X, y=None):
        """Creates forecast for dates specified in X

        Parameters
        ----------
        X : pd.DataFrame with column (self.time_col_), the dates for future prediction. Other columns are ignored.
            timestamps to make forecast

        y : ignored

        Returns
        -------
        df : pd.DataFrame with shape = (X.shape[0], 2)
            Forecasted values for dates in X
        """
        x_array = X[[self.time_col_]].to_numpy()
        predictions = self.model.predict(X=x_array)

        df = pd.DataFrame({
            TIME_COL: X[self.time_col_],
            PREDICTED_COL: predictions
        })

        return df

    def summary(self):
        """Prints input parameters and DummyRegressor model parameters
        """
        log_message(self, LoggingLevelEnum.DEBUG)
        log_message(self.model, LoggingLevelEnum.DEBUG)

    def score(self, X, y, sample_weight=None):
        """Default scorer for the estimator. Returns error based on ``score_func``.
        Because this is often used as null model, it is more informative to return ``score_func`` of the null model
        rather than R2_null_model_score of null model against something else"""
        y_pred = self.predict(X)[PREDICTED_COL]
        return self.score_func(y, y_pred)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen, Sayan Patra, Rachit Kumar
"""Common python utility functions."""
import copy
import dataclasses
import functools
import math
import re
import warnings
from dataclasses import field

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from pandas.testing import assert_index_equal
from pandas.testing import assert_series_equal


def update_dictionary(default_dict, overwrite_dict=None, allow_unknown_keys=True):
    """Adds default key-value pairs to items in ``overwrite_dict``.

    Merges the items in ``default_dict`` and ``overwrite_dict``,
    preferring ``overwrite_dict`` if there are conflicts.

    Parameters
    ----------
    default_dict: `dict`
        Dictionary of default values.
    overwrite_dict: `dict` or None, optional, default None
        User-provided dictionary that overrides the defaults.
    allow_unknown_keys: `bool`, optional, default True
        If false, raises an error if ``overwrite_dict`` contains a key that is
        not in ``default_dict``.

    Raises
    ------
    ValueError
        if ``allow_unknown_keys`` is False and ``overwrite_dict``
        has keys that are not in ``default_dict``.

    Returns
    -------
    updated_dict : `dict`
        Updated dictionary.
        Returns ``overwrite_dicts``, with default values added
        based on ``default_dict``.
    """
    if overwrite_dict is None:
        overwrite_dict = {}

    if not allow_unknown_keys:
        extra_keys = overwrite_dict.keys() - default_dict.keys()
        if extra_keys:
            raise ValueError(f"Unexpected key(s) found: {extra_keys}. "
                             f"The valid keys are: {default_dict.keys()}")

    return dict(default_dict, **overwrite_dict)


def update_dictionaries(default_dict, overwrite_dicts=None, allow_unknown_keys=True):
    """Adds default key-value pairs to items in ``overwrite_dicts``.

    Merges the items in ``default_dict`` and ``overwrite_dicts``,
    preferring ``overwrite_dict`` if there are conflicts.

    If ``overwrite_dicts`` is a list of dictionaries, the merge is
    applied to each dictionary in the list.

    Parameters
    ----------
    default_dict: `dict`
        Dictionary of default values.
    overwrite_dicts: `dict` or None or `list` [`dict` or None], optional, default None
        User-provided dictionary that overrides the defaults,
        or a list of such dictionaries.
    allow_unknown_keys: `bool`, optional, default True
        If false, raises an error if ``overwrite_dicts`` contains a key that is
        not in ``default_dict``.

    Returns
    -------
    updated_dict : `dict` or `list` [`dict`]
        Updated dictionary of list of dictionaries.
        Returns ``overwrite_dicts``, with default values added
        to each dictionary based on ``default_dict``.
    """
    if isinstance(overwrite_dicts, (list, tuple)):
        updated_dict = [
            update_dictionary(
                default_dict,
                overwrite_dict=item,
                allow_unknown_keys=allow_unknown_keys)
            for item in overwrite_dicts]
    else:
        updated_dict = update_dictionary(
            default_dict,
            overwrite_dict=overwrite_dicts,
            allow_unknown_keys=allow_unknown_keys)
    return updated_dict


def unique_elements_in_list(array):
    """Returns the unique elements in the input list,
        preserving the original order.

    Parameters
    ----------
    array: `List` [`any`]
        List of elements.

    Returns
    -------
    unique_array : `List` [`any`]
        Unique elements in `array`, preserving the order of first appearance.
    """
    seen = set()
    result = []
    for item in array:
        if item not in seen:
            result.append(item)
            seen.add(item)
    return result


def unique_dict_in_list(array):
    """Returns the unique dictionaries in the input list,
    preserving the original order. Replaces ``unique_elements_in_list``
    because `dict` is not hashable.

    Parameters
    ----------
    array: `List` [`dict`]
        List of dictionaries.

    Returns
    -------
    unique_array : `List` [`dict`]
        Unique dictionaries in `array`, preserving the order of first appearance.
    """
    if not array:
        return array
    result = []
    # Brute force.
    # Avoids comparing json dumped ordered dictionary.
    # The reason is that when dictionary contains list/dict unhashable items, set does not work.
    for item in array:
        if not any([item == element for element in result]):
            result.append(item)
    return result


def get_pattern_cols(cols, pos_pattern=None, neg_pattern=None):
    """Get columns names from a list that matches ``pos_pattern``,
    but does not match ``neg_pattern``.

    If a column name matches both ``pos_pattern`` and ``neg_pattern``, it is excluded.

    Parameters
    ----------
    cols : `List` ['str']
        Usually column names of a DataFrame.
    pos_pattern : regular expression
        If column name matches this pattern, it is included in the output.
    neg_pattern : regular expression
        If column name matches this pattern, it is excluded from the output.

    Returns
    -------
    pattern_cols : `List` ['str']
        List of column names that match the pattern.
    """
    if pos_pattern is None:
        pos_pattern_cols = []
    else:
        pos_regex = re.compile(pos_pattern)
        pos_pattern_cols = [col for col in cols if pos_regex.findall(col)]
    if neg_pattern is None:
        neg_pattern_cols = []
    else:
        neg_regex = re.compile(neg_pattern)
        neg_pattern_cols = [col for col in cols if neg_regex.findall(col)]

    pattern_cols = [col for col in cols if
                    col in pos_pattern_cols and col not in neg_pattern_cols]

    return pattern_cols


def assert_equal(
        actual,
        expected,
        ignore_list_order=False,
        rel=1e-5,
        dict_path="",
        ignore_keys=None,
        **kwargs):
    """Generic equality function that raises an ``AssertionError`` if the objects are not equal.

    Notes
    -----
    Works with pandas.DataFrame, pandas.Series, numpy.ndarray, str, int, float,
    bool, None, or a dictionary or list of such items, with arbitrary nesting
    of dictionaries and lists.

    Does not check equivalence of functions, or work with nested numpy arrays.

    Parameters
    ----------
    actual : `pandas.DataFrame`, `pandas.Series`, `numpy.array`, `str`, `int`,
     `float`, `bool`, `None`, or a dictionary or list of such items
        Actual value.
    expected : `pandas.DataFrame`, `pandas.Series`, `numpy.array`, `str`, `int`,
     `float`, `bool`, `None`, or a dictionary or list of such items
        Expected value to compare against.
    ignore_list_order : `bool`, optional, default False
        If True, lists are considered equal if they contain the same elements. This option is valid
        only if the list can be sorted (all elements can be compared to each other).
        If False, lists are considered equal if they contain the same elements in the same order.
    rel : `float`, optional, default 1e-5
        To check int and float, passed to ``rel`` argument of `pytest.approx`.
        To check numpy arrays, passed to ``rtol`` argument of
        `numpy.testing.assert_allclose`.
        To check pandas dataframe, series, and index, passed to ``rtol`` argument of
        `pandas.testing.assert_frame_equal`, `pandas.testing.assert_series_equal`,
        `pandas.testing.assert_index_equal`.
    dict_path : `str`, optional, default ""
        Location within nested dictionary of the original call to this function.
        User should not set this parameter.
    ignore_keys : `dict`, optional, default None
        Keys to ignore in equality comparison. This only applies if
        `expected` is a dictionary.
        Does not compare the values of these keys. However,
        still returns false if the key is not present.

        Can be a nested dictionary. Terminal keys are those whose values should
        not be compared.


        If the expected value is an nested dictionary, this dictionary can
        also be nested, with the same structure.
        For example, if expected:
        expected = {
            "k1": {
                "k1": 1,
                "k2": [1, 2, 3],
            }
            "k2": {
                "k1": "abc"
            }
        }
        Then the following ``ignore_keys`` will ignore
        dict["k1"]["k1"] and dict["k2"]["k1"] in the comparison.
        ignore_keys = {
            "k1": {
                "k1": False  # The value can be anything, the keys determine what's ignored
            },
            "k2": {
                "k1": "skip"  # The value can be anything, the keys determine what's ignored
            }
        }
        This ``ignore_keys`` will ignore
        dict["k1"] and dict["k2"]["k1"] in the comparison. "k1" is
        ignored entirely because its value is not a dictionary.
        ignore_keys = {
            "k1": None  # The value can be anything, the keys determine what's ignored
            "k2": {
                "k1": None
            }
        }
    kwargs : keyword args, optional
        Keyword args to pass to `pandas.util.testing.assert_frame_equal`,
        `pandas.util.testing.assert_series_equal`.

    Raises
    ------
    AssertionError
        If actual does not match expected.
    """
    # a message to add to all error messages
    location = f"dictionary location: {dict_path}"
    message = "" if dict_path == "" else f"Error at {location}.\n"

    if expected is None:
        if actual is not None:
            raise AssertionError(f"{message}Actual should be None, found {actual}.")
    elif isinstance(expected, pd.DataFrame):
        if not isinstance(actual, pd.DataFrame):
            raise AssertionError(f"{message}Actual should be a pandas DataFrame, found {actual}.")
        # leverages pandas assert function and add `message` to the error
        try:
            assert_frame_equal(
                actual,
                expected,
                rtol=rel,
                **kwargs)
        except AssertionError as e:
            import sys
            raise type(e)(f"{e}{message}").with_traceback(sys.exc_info()[2])
    elif isinstance(expected, pd.Series):
        if not isinstance(actual, pd.Series):
            raise AssertionError(f"{message}Actual should be a pandas Series, found {actual}.")
        try:
            assert_series_equal(
                actual,
                expected,
                rtol=rel,
                **kwargs)
        except AssertionError as e:
            import sys
            raise type(e)(f"{message}{e}").with_traceback(sys.exc_info()[2])
    elif isinstance(expected, pd.Index):
        if not isinstance(actual, pd.Index):
            raise AssertionError(f"{message}Actual should be a pandas Index, found {actual}.")
        try:
            assert_index_equal(
                actual,
                expected,
                rtol=rel,
                **kwargs)
        except AssertionError as e:
            import sys
            raise type(e)(f"{message}{e}").with_traceback(sys.exc_info()[2])
    elif isinstance(expected, np.ndarray):
        if not isinstance(actual, np.ndarray):
            raise AssertionError(f"{message}Actual should be a numpy array, found {actual}.")
        np.testing.assert_allclose(
            actual,
            expected,
            rtol=rel,
            err_msg=message)
    elif isinstance(expected, (list, tuple)):
        if not isinstance(actual, (list, tuple)):
            raise AssertionError(f"{message}Actual should be a list or tuple, found {actual}.")
        if not len(actual) == len(expected):
            raise AssertionError(f"{message}Lists have different length. "
                                 f"Actual: {actual}. Expected: {expected}.")
        if ignore_list_order:
            # order doesn't matter
            actual = sorted(actual)
            expected = sorted(expected)
        # element-wise comparison
        if ignore_keys is not None:
            warnings.warn(f"At {location}. `ignore_keys` is {ignore_keys}, but found a list. "
                          f"No keys will be ignored")
        for (item_actual, item_expected) in zip(actual, expected):
            assert_equal(
                item_actual,
                item_expected,
                ignore_list_order=ignore_list_order,
                rel=rel,
                dict_path=dict_path,
                ignore_keys=None,
                **kwargs)
    elif isinstance(expected, dict):
        # dictionaries are equal if their keys and values are equal
        if not isinstance(actual, dict):
            raise AssertionError(f"{message}Actual should be a dict, found {actual}.")
        # checks the keys
        if not actual.keys() == expected.keys():
            raise AssertionError(f"{message}Dict keys do not match. "
                                 f"Actual: {actual.keys()}. Expected: {expected.keys()}.")
        # check the next level of nesting, if not ignored by `ignore_keys`
        for k, expected_item in expected.items():
            if ignore_keys is not None and k in ignore_keys.keys():
                if isinstance(ignore_keys[k], dict):
                    # specific keys within the value are ignored
                    new_ignore_keys = ignore_keys[k]
                else:
                    # the entire value is ignored
                    continue
            else:
                # the key is not ignored, so its value should be fully compared
                new_ignore_keys = None

            # appends the key to the path
            new_path = f"{dict_path}['{k}']" if dict_path != "" else f"dict['{k}']"
            assert_equal(
                actual[k],
                expected_item,
                ignore_list_order=ignore_list_order,
                rel=rel,
                dict_path=new_path,
                ignore_keys=new_ignore_keys,
                **kwargs)
    elif isinstance(expected, (int, float)):
        if not isinstance(actual, (int, float)):
            raise AssertionError(f"{message}Actual should be numeric, found {actual}.")
        if not math.isclose(actual, expected, rel_tol=rel, abs_tol=0.0):
            raise AssertionError(f"{message}Actual does not match expected. "
                                 f"Actual: {actual}. Expected: {expected}.")
    else:
        if actual != expected:
            raise AssertionError(f"{message}Actual does not match expected. "
                                 f"Actual: {actual}. Expected: {expected}.")


def dictionary_values_to_lists(hyperparameter_dict, hyperparameters_list_type=None):
    """Given a dictionary, returns a copy whose values are
     either lists, distributions with a ``rvs`` method, or None.

    The output is suitable for hyperparameter grid search.

    Does this by converting values that do not conform into
    singleton elements inside a list.

    Parameters
    ----------
    hyperparameter_dict : `dict` [`str`, `any`]
        Dictionary of hyperparameters.

    hyperparameters_list_type : `set` [`str`] or `dict` [`str`, `list`] or None, optional, default None
        Hyperparameters that must be a `list` or other recognized value.
        e.g. ``regressor_cols`` is `list` or None, ``holiday_lookup_countries`` is `list` or "auto"
        or None.

        Thus, a flat list must become nested. E.g. ["US", "UK"] must be converted to [["US", "UK"]].

        Specifically, the values in ``hyperparameter_dict`` must be of type
        `list` [`list` or other accepted value].

            * If a set, other accepted value = [None]
            * If a dict, other accepted value is specified by the key's value, a
              list of valid options.

              For example, to allow `list` or "auto" or None, use
              ``hyperparameters_list_type={"regressor_cols": [None, "auto"]}``.

              For example, ``hyperparameters_list_type={"regressor_cols": [None]}`` is equivalent to
              ``hyperparameters_list_type={"regressor_cols"}`` using the set specification.

    Notes
    -----
    These values are unchanged:

        * [None]
        * ["value1", "value2"]
        * [[1, 2], [3, 4]]
        * [[1], None, [3]]
        * scipy.stats.expon(scale=.1)

    These values are put in a list:

        * None
        * 1
        * np.array([1, 2, 3])
        * {"k": "v"}

    These values are put in a list if their key is
    in ``hyperparameters_list_type`` and the other
    acceptable value is [None], otherwise unchanged:
        * []
        * [1, 2, 3]
        * (1, 2, 3)
        * ["value1", "value2"]

    Raises
    ------
    ValueError
        If a key is in ``hyperparameters_list_type`` but
        its value is not a list, tuple, or None.

    Returns
    -------
    hyperparameter_grid : `dict` [`str`, `list` [`any`] or distribution with ``rvs`` method]
        A dictionary suitable to pass as ``param_distributions`` to
        `sklearn.model_selection.RandomizedSearchCV` for grid search.

        As explained in `sklearn.model_selection.RandomizedSearchCV`:
        Dictionary with parameters names (string) as keys and distributions
        or lists of parameters to try. Distributions must provide a ``rvs``
        method for sampling (such as those from scipy.stats.distributions).
        If a list is given, it is sampled uniformly.
    """
    hyperparameter_grid = hyperparameter_dict.copy()
    if hyperparameters_list_type is None:
        hyperparameters_list_type = {}

    for param, value in hyperparameter_dict.items():
        # Distribution is allowed
        is_distribution = hasattr(value, "rvs")
        is_list = isinstance(value, (list, tuple))

        if param in hyperparameters_list_type:
            if isinstance(hyperparameters_list_type, dict):
                recognized_values = hyperparameters_list_type.get(param, [None])
            else:
                recognized_values = [None]

            if is_list:
                # A list is provided
                if (len(value) == 0 or
                        not all([isinstance(list_item, (list, tuple))
                                 or list_item in recognized_values
                                 for list_item in value])):
                    # Not all its values are acceptable for the hyperparameter,
                    # therefore enclose the value in a list.
                    hyperparameter_grid[param] = [value]
            elif value in recognized_values:
                # Not a list, but the value is acceptable
                hyperparameter_grid[param] = [value]
            else:
                raise ValueError(
                    f"The value for {param} must be a list, tuple, or one of {recognized_values}, "
                    f"found {value}.")

        else:
            # Any list or distribution is allowed.
            # Violating elements are enclosed the value in a list
            if not is_list and not is_distribution:
                hyperparameter_grid[param] = [value]

    return hyperparameter_grid


def dictionaries_values_to_lists(hyperparameter_dicts, hyperparameters_list_type=None):
    """Calls `~greykite.common.utils.python_utils.dictionary_values_to_lists`
    on the provided dictionary or on each item in a list of dictionaries.

    ``dictionary_values_to_lists`` returns a copy whose values are
     either lists, distributions with a ``rvs`` method, or None.

    Parameters
    ----------
    hyperparameter_dicts : `dict` [`str`, `any`] or `list` [`dict` [`str`, `any`]]
        Dictionary of hyperparameters, or list of such dictionaries

    hyperparameters_list_type : `set` [`str`] or `dict` [`str`, `list`] or None, optional, default None
        Hyperparameters that must be a `list` or other recognized value.
        e.g. ``regressor_cols`` is `list` or None, ``holiday_lookup_countries`` is `list` or "auto"
        or None.

        Thus, a flat list must become nested. E.g. ["US", "UK"] must be converted to [["US", "UK"]].

        Specifically, the values in ``hyperparameter_dict`` must be of type
        `list` [`list` or other accepted value].

            * If a set, other accepted value = [None]
            * If a dict, other accepted value is specified by the key's value, a
              list of valid options.

              For example, to allow `list` or "auto" or None, use
              ``hyperparameters_list_type={"regressor_cols": [None, "auto"]}``.

              For example, ``hyperparameters_list_type={"regressor_cols": [None]}`` is equivalent to
              ``hyperparameters_list_type={"regressor_cols"}`` using the set specification.

    Returns
    -------
    hyperparameter_grid : `dict` [`str`, `list` [`any`], or distribution with ``rvs`` method] or `list` [`dict]
        A dictionary or list of dictionaries suitable to pass as
        ``param_distributions`` to `sklearn.model_selection.RandomizedSearchCV`
        for grid search.

        As explained in `sklearn.model_selection.RandomizedSearchCV`:
        Dictionary with parameters names (string) as keys and distributions
        or lists of parameters to try. Distributions must provide a ``rvs``
        method for sampling (such as those from scipy.stats.distributions).
        If a list is given, it is sampled uniformly.

    See Also
    --------
    `~greykite.common.utils.python_utils.dictionary_values_to_lists`
    """
    if isinstance(hyperparameter_dicts, (list, tuple)):
        hyperparameter_grids = [
            dictionary_values_to_lists(
                item,
                hyperparameters_list_type=hyperparameters_list_type)
            for item in hyperparameter_dicts]
    else:
        hyperparameter_grids = dictionary_values_to_lists(
            hyperparameter_dicts,
            hyperparameters_list_type=hyperparameters_list_type)
    return hyperparameter_grids


def unique_in_list(array, ignored_elements=()):
    """Returns unique elements in ``array``, removing
    all levels of nesting if found.

    Parameters
    ----------
    array : `list` [`any`] or None
        List of items, with arbitrary level of nesting.
    ignored_elements : `tuple`, default ()
        Elements not to include in the output

    Returns
    -------
    unique_elements : `list`
        Unique elements in array, ignoring up to
        `level` levels of nesting.
        Elements that are `None` are removed from the output.
    """
    unique_elements = set()
    if array is not None:
        for item in array:
            if isinstance(item, (list, tuple)):
                unique_in_item = unique_in_list(item, ignored_elements=ignored_elements)
                unique_in_item = set(unique_in_item) if unique_in_item is not None else {}
                unique_elements.update(unique_in_item)
            elif item not in ignored_elements:
                unique_elements.add(item)
    return list(unique_elements) if unique_elements else None


def flatten_list(array):
    """Flattens an array by removing 1 level of nesting.

    Parameters
    ----------
    array : `list` [`list`]
        List of lists.

    Returns
    -------
    flat_arr : `list`
        Removes one level of nesting from the array.
        [[4], [3, 2], [1, [0]]] becomes [4, 3, 2, 1, [0]].
    """
    return [item for sublist in array for item in sublist]


def reorder_columns(df, order_dict=None):
    """Orders columns according to ``order_dict``.

    Can be used to order columns according to hierarchical
    constraints. Consider the tree where a parent is the sum
    of its children. Let a node's label be its BFS traversal order,
    with the root as 0. Use ``order_dict`` to map column names
    to these node labels, to get the dataframe in BFS traversal order,
    matching the structure of the tree.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Input data frame.
    order_dict : `dict` [`str`, `float`] or None
        How to order the columns.
        The key is the column name, the value is its position.
        Columns are returned in ascending order by value from left to right.
        Only column specified by ``order_dict`` are included in the output.
        If None, returns the original ``df``.

    Returns
    -------
    reordered_df : `pandas.DataFrame`
        ``df`` with the selected columns reordered.
    """
    if order_dict is not None:
        order_tuples = list(order_dict.items())
        order_tuples = sorted(order_tuples, key=lambda x: x[1])
        order_names = [x[0] for x in order_tuples]
        df = df[order_names]
    return df


def apply_func_to_columns(row_func, cols):
    """Returns a function that applies ``row_func`` to
    the selected ``cols``. Helper function for
    `~greykite.framework.output.univariate_forecast.UnivariateForecast.autocomplete_map_func_dict`.

    Parameters
    ----------
    row_func : callable
        A function.
    cols : `list` [`str` or `int`]
        Names of the columns (or dictionary keys, list indices)
        to pass to ``row_func``.

    Returns
    -------
    new_func : callable
        Takes ``row`` and returns the result of ``row_func``
        applied to the selected values ``row[col]``.
    """
    def new_func(row):
        return row_func(*[row[col] for col in cols])
    return new_func


def get_integer(val=None, name="value", min_value=0, default_value=0):
    """Returns integer value from input, with basic validation

    Parameters
    ----------
    val : `float` or None, default None
        Value to convert to integer.
    name : `str`, default "value"
        What the value represents.
    min_value : `float`, default 0
        Minimum allowed value.
    default_value : `float` , default 0
        Value to be used if ``val`` is None.

    Returns
    -------
    val : `int`
        Value parsed as an integer.
    """
    if val is None:
        val = default_value
    try:
        orig = val
        val = int(val)
    except ValueError:
        raise ValueError(f"{name} must be an integer")
    else:
        if val != orig:
            warnings.warn(f"{name} converted to integer {val} from {orig}")

    if not val >= min_value:
        raise ValueError(f"{name} must be >= {min_value}")

    return val


def mutable_field(mutable_default_value) -> dataclasses.field:
    """Can be used to set the default value in a dataclass
    to a mutable value.

    Provides a factory function that returns a copy of the provided argument.

    Parameters
    ----------
    mutable_default_value : Any
        The default value to use for the field.

    Returns
    -------
    field : `dataclasses.field`
        Set the default value to this value.

    Examples
    --------
    >>> from dataclasses import dataclass
    >>> from typing import List
    >>> from greykite.common.python_utils import mutable_field
    >>> @dataclass
    >>> class D:
    >>>     x: List = mutable_field([1, 2, 3])
    >>>
    >>> assert D().x is not D().x
    >>> assert D().x == [1, 2, 3]
    """
    return field(default_factory=lambda: copy.deepcopy(mutable_default_value))


def ignore_warnings(category):
    """Returns a decorator to ignore all warnings
    in the specified category.

    Parameters
    ----------
    category : class
        Any warning that is a subclass of this category is ignored.

    Returns
    -------
    decorator_ignore : function
        A decorator that ignores all warnings in the category.
    """
    def decorator_ignore(fn):
        @functools.wraps(fn)
        def fn_ignore(*args, **kwargs):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category)
                return fn(*args, **kwargs)
        return fn_ignore
    return decorator_ignore



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Template for `greykite.sklearn.estimator.simple_silverkite_estimator`.
Takes input data and forecast config,
and returns parameters to call
:func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
"""

import dataclasses
import warnings
from typing import Type

from greykite.algo.forecast.silverkite.constants.silverkite_holiday import SilverkiteHoliday
from greykite.common.python_utils import dictionaries_values_to_lists
from greykite.common.python_utils import unique_dict_in_list
from greykite.common.python_utils import unique_in_list
from greykite.common.python_utils import update_dictionaries
from greykite.framework.templates.autogen.forecast_config import ModelComponentsParam
from greykite.framework.templates.base_template import BaseTemplate
from greykite.framework.templates.simple_silverkite_template_config import SimpleSilverkiteTemplateConstants
from greykite.sklearn.estimator.base_forecast_estimator import BaseForecastEstimator
from greykite.sklearn.estimator.simple_silverkite_estimator import SimpleSilverkiteEstimator


class SimpleSilverkiteTemplate(BaseTemplate):
    """A template for :class:`~greykite.sklearn.estimator.simple_silverkite_estimator.SimpleSilverkiteEstimator`.

    Takes input data and optional configuration parameters
    to customize the model. Returns a set of parameters to call
    :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.

    Notes
    -----
    The attributes of a `~greykite.framework.templates.autogen.forecast_config.ForecastConfig` for
    :class:`~greykite.sklearn.estimator.simple_silverkite_estimator.SimpleSilverkiteEstimator` are:

        computation_param: `ComputationParam` or None, default None
            How to compute the result. See
            :class:`~greykite.framework.templates.autogen.forecast_config.ComputationParam`.
        coverage: `float` or None, default None
            Intended coverage of the prediction bands (0.0 to 1.0).
            Same as coverage in ``forecast_pipeline``.
            You may tune how the uncertainty is computed
            via `model_components.uncertainty["uncertainty_dict"]`.
        evaluation_metric_param: `EvaluationMetricParam` or None, default None
            What metrics to evaluate. See
            :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationMetricParam`.
        evaluation_period_param: `EvaluationPeriodParam` or None, default None
            How to split data for evaluation. See
            :class:`~greykite.framework.templates.autogen.forecast_config.EvaluationPeriodParam`.
        forecast_horizon: `int` or None, default None
            Number of periods to forecast into the future. Must be > 0
            If None, default is determined from input data frequency
            Same as forecast_horizon in `forecast_pipeline`
        metadata_param: `MetadataParam` or None, default None
            Information about the input data. See
            :class:`~greykite.framework.templates.autogen.forecast_config.MetadataParam`.
        model_components_param: `ModelComponentsParam`, `list` [`ModelComponentsParam`] or None, default None
            Parameters to tune the model. See
            :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`.
            The fields are dictionaries with the following items.

            See inline comments on which values accept lists for grid search.

            seasonality: `dict` [`str`, `any`] or None, optional
                Seasonality configuration dictionary, with the following optional keys.
                (keys are SilverkiteSeasonalityEnum members in lower case).

                The keys are parameters of
                `~greykite.algo.forecast.silverkite.forecast_simple_silverkite.forecast_simple_silverkite`.
                Refer to that function for more details.

                ``"yearly_seasonality"``: `str` or `bool` or `int` or a list of such values for grid search, default 'auto'
                    Determines the yearly seasonality
                    'auto', True, False, or a number for the Fourier order
                ``"quarterly_seasonality"``: `str` or `bool` or `int` or a list of such values for grid search, default 'auto'
                    Determines the quarterly seasonality
                    'auto', True, False, or a number for the Fourier order
                ``"monthly_seasonality"``: `str` or `bool` or `int` or a list of such values for grid search, default 'auto'
                    Determines the monthly seasonality
                    'auto', True, False, or a number for the Fourier order
                ``"weekly_seasonality"``: `str` or `bool` or `int` or a list of such values for grid search, default 'auto'
                    Determines the weekly seasonality
                    'auto', True, False, or a number for the Fourier order
                ``"daily_seasonality"``: `str` or `bool` or `int` or a list of such values for grid search, default 'auto'
                    Determines the daily seasonality
                    'auto', True, False, or a number for the Fourier order

            growth: `dict` [`str`, `any`] or None, optional
                Growth configuration dictionary with the following optional key:

                ``"growth_term"``: `str` or None or a list of such values for grid search
                    How to model the growth.
                    Valid options are "linear", "quadratic", "sqrt", "cubic", "cuberoot"
                    All these terms have their origin at the train start date.

            events: `dict` [`str`, `any`] or None, optional
                Holiday/events configuration dictionary with the following optional keys:

            ``"holiday_lookup_countries"``: `list` [`str`] or "auto" or None or a list of such values for grid search, default "auto"
                The countries that contain the holidays you intend to model
                (``holidays_to_model_separately``).

                * If "auto", uses a default list of countries
                  that contain the default ``holidays_to_model_separately``.
                  See `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.HOLIDAY_LOOKUP_COUNTRIES_AUTO`.
                * If a list, must be a list of country names.
                * If None or an empty list, no holidays are modeled.

            ``"holidays_to_model_separately"``: `list` [`str`] or "auto" or `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.ALL_HOLIDAYS_IN_COUNTRIES` or None or a list of such values for grid search, default "auto"  # noqa: E501
                Which holidays to include in the model.
                The model creates a separate key, value for each item in ``holidays_to_model_separately``.
                The other holidays in the countries are grouped together as a single effect.

                * If "auto", uses a default list of important holidays.
                  See `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.HOLIDAYS_TO_MODEL_SEPARATELY_AUTO`.
                * If `~greykite.algo.forecast.silverkite.constants.silverkite_holiday.SilverkiteHoliday.ALL_HOLIDAYS_IN_COUNTRIES`,
                  uses all available holidays in ``holiday_lookup_countries``. This can often
                  create a model that has too many parameters, and should typically be avoided.
                * If a list, must be a list of holiday names.
                * If None or an empty list, all holidays in ``holiday_lookup_countries`` are grouped together
                  as a single effect.

                Use ``holiday_lookup_countries`` to provide a list of countries where these holiday occur.

                ``"holiday_pre_num_days"``: `int` or a list of such values for grid search, default 2
                    model holiday effects for pre_num days before the holiday.
                    The unit is days, not periods. It does not depend on input data frequency.
                ``"holiday_post_num_days"``: `int` or a list of such values for grid search, default 2
                    model holiday effects for post_num days after the holiday.
                    The unit is days, not periods. It does not depend on input data frequency.

            ``"holiday_pre_post_num_dict"``: `dict` [`str`, (`int`, `int`)] or None, default None
                Overrides ``pre_num`` and ``post_num`` for each holiday in
                ``holidays_to_model_separately``.
                For example, if ``holidays_to_model_separately`` contains "Thanksgiving" and "Labor Day",
                this parameter can be set to ``{"Thanksgiving": [1, 3], "Labor Day": [1, 2]}``,
                denoting that the "Thanksgiving" ``pre_num`` is 1 and ``post_num`` is 3, and "Labor Day"
                ``pre_num`` is 1 and ``post_num`` is 2.
                Holidays not specified use the default given by ``pre_num`` and ``post_num``.
            ``"daily_event_df_dict"``: `dict` [`str`, `pandas.DataFrame`] or None, default None
                A dictionary of data frames, each representing events data for the corresponding key.
                Specifies additional events to include besides the holidays specified above. The format
                is the same as in `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
                The DataFrame has two columns:

                    - The first column contains event dates. Must be in a format
                      recognized by `pandas.to_datetime`. Must be at daily
                      frequency for proper join. It is joined against the time
                      in ``df``, converted to a day:
                      ``pd.to_datetime(pd.DatetimeIndex(df[time_col]).date)``.
                    - the second column contains the event label for each date

                The column order is important; column names are ignored.
                The event dates must span their occurrences in both the training
                and future prediction period.

                During modeling, each key in the dictionary is mapped to a categorical variable
                named ``f"{EVENT_PREFIX}_{key}"``, whose value at each timestamp is specified
                by the corresponding DataFrame.

                For example, to manually specify a yearly event on September 1
                during a training/forecast period that spans 2020-2022::

                    daily_event_df_dict = {
                        "custom_event": pd.DataFrame({
                            "date": ["2020-09-01", "2021-09-01", "2022-09-01"],
                            "label": ["is_event", "is_event", "is_event"]
                        })
                    }

                It's possible to specify multiple events in the same df. Two events,
                ``"sep"`` and ``"oct"`` are specified below for 2020-2021::

                    daily_event_df_dict = {
                        "custom_event": pd.DataFrame({
                            "date": ["2020-09-01", "2020-10-01", "2021-09-01", "2021-10-01"],
                            "event_name": ["sep", "oct", "sep", "oct"]
                        })
                    }

                Use multiple keys if two events may fall on the same date. These events
                must be in separate DataFrames::

                    daily_event_df_dict = {
                        "fixed_event": pd.DataFrame({
                            "date": ["2020-09-01", "2021-09-01", "2022-09-01"],
                            "event_name": "fixed_event"
                        }),
                        "moving_event": pd.DataFrame({
                            "date": ["2020-09-01", "2021-08-28", "2022-09-03"],
                            "event_name": "moving_event"
                        }),
                    }

                The multiple event specification can be used even if events never overlap. An
                equivalent specification to the second example::

                    daily_event_df_dict = {
                        "sep": pd.DataFrame({
                            "date": ["2020-09-01", "2021-09-01"],
                            "event_name": "is_event"
                        }),
                        "oct": pd.DataFrame({
                            "date": ["2020-10-01", "2021-10-01"],
                            "event_name": "is_event"
                        }),
                    }

                Note: All these events are automatically added to the model. There is no need
                to specify them in ``extra_pred_cols`` as you would for
                `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.

                Note: Do not use `~greykite.common.constants.EVENT_DEFAULT`
                in the second column. This is reserved to indicate dates that do not
                correspond to an event.
            changepoints: `dict` [`str`, `dict`] or None, optional
                Specifies the changepoint configuration. Dictionary with the following
                optional key:

                ``"changepoints_dict"``: `dict` or None or a list of such values for grid search
                    Changepoints dictionary passed to ``forecast_simple_silverkite``. A dictionary
                    with the following optional keys:

                    ``"method"``: `str`
                        The method to locate changepoints. Valid options:

                            - "uniform". Places n_changepoints evenly spaced changepoints to allow growth to change.
                            - "custom". Places changepoints at the specified dates.
                            - "auto". Automatically detects change points.

                        Additional keys to provide parameters for each particular method are described below.
                    ``"continuous_time_col"``: `str` or None
                        Column to apply `growth_func` to, to generate changepoint features
                        Typically, this should match the growth term in the model
                    ``"growth_func"``: callable or None
                        Growth function (`numeric` -> `numeric`). Changepoint features are created
                        by applying `growth_func` to "continuous_time_col" with offsets.
                        If None, uses identity function to use `continuous_time_col` directly
                        as growth term

                    If changepoints_dict["method"] == "uniform", this other key is required:

                        ``"n_changepoints"``: `int`
                            number of changepoints to evenly space across training period

                    If changepoints_dict["method"] == "custom", this other key is required:

                        ``"dates"``: `list` [`int` or `float` or `str` or `datetime`]
                            Changepoint dates. Must be parsable by pd.to_datetime.
                            Changepoints are set at the closest time on or after these dates
                            in the dataset.

                    If changepoints_dict["method"] == "auto", optional keys can be passed that match the parameters in
                    `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_trend_changepoints`
                    (except ``df``, ``time_col`` and ``value_col``, which are already known).
                    To add manually specified changepoints to the automatically detected ones, the keys ``dates``,
                    ``combine_changepoint_min_distance`` and ``keep_detected`` can be specified, which correspond to the
                    three parameters ``custom_changepoint_dates``, ``min_distance`` and ``keep_detected`` in
                    `~greykite.algo.changepoint.adalasso.changepoints_utils.combine_detected_and_custom_trend_changepoints`.

                ``"seasonality_changepoints_dict"``: `dict` or None or a list of such values for grid search
                        seasonality changepoints dictionary passed to ``forecast_simple_silverkite``. The optional
                        keys are the parameters in
                        `~greykite.algo.changepoint.adalasso.changepoint_detector.ChangepointDetector.find_seasonality_changepoints`.
                        You don't need to provide ``df``, ``time_col``, ``value_col`` or ``trend_changepoints``, since they
                        are passed with the class automatically.

            autoregression: `dict` [`str`, `dict`] or None, optional
                Specifies the autoregression configuration. Dictionary with the following optional key:

                ``"autoreg_dict"``: `dict` or None or a list of such values for grid search
                    A dictionary with arguments for `~greykite.common.features.timeseries_lags.build_autoreg_df`.
                    That function's parameter ``value_col`` is inferred from the input of
                    current function ``forecast_silverkite``. Other keys are:

                        * ``"lag_dict"`` : `dict` or None
                        * ``"agg_lag_dict"`` : `dict` or None
                        * ``"series_na_fill_func"`` : callable

                    See more details for above parameters in
                    `~greykite.common.features.timeseries_lags.build_autoreg_df`.

            regressors: `dict` [`str`, `any`] or None, optional
                Specifies the regressors to include in the model (e.g. macro-economic factors).
                Dictionary with the following optional keys:

                ``"regressor_cols"``: `list` [`str`] or None or a list of such values for grid search
                    The columns in ``df`` to use as regressors.
                    Note that regressor values must be available in ``df`` for all prediction dates.
                    Thus, ``df`` will contain timestamps for both training and future prediction.

                    * regressors must be available on all dates
                    * the response must be available for training dates (metadata["value_col"])

                Use ``extra_pred_cols`` to specify interactions of any model terms with the regressors.

            uncertainty: `dict` [`str`, `dict`] or None, optional
                Along with ``coverage``, specifies the uncertainty interval configuration. Use ``coverage``
                to set interval size. Use ``uncertainty`` to tune the calculation.

                ``"uncertainty_dict"``: `str` or `dict` or None or a list of such values for grid search
                    "auto" or a dictionary on how to fit the uncertainty model.
                    If a dictionary, valid keys are:

                        ``"uncertainty_method"``: `str`
                            The title of the method.
                            Only ``"simple_conditional_residuals"`` is implemented
                            in `~greykite.algo.common.ml_models.fit_ml_model`
                            which calculates intervals using residuals.

                        ``"params"``: `dict`
                            A dictionary of parameters needed for
                            the requested ``uncertainty_method``. For example, for
                            ``uncertainty_method="simple_conditional_residuals"``, see
                            parameters of `~greykite.algo.uncertainty.conditional.conf_interval.conf_interval`:

                                * ``"conditional_cols"``
                                * ``"quantiles"``
                                * ``"quantile_estimation_method"``
                                * ``"sample_size_thresh"``
                                * ``"small_sample_size_method"``
                                * ``"small_sample_size_quantile"``

                            The default value for ``quantiles`` is inferred from coverage.

                    If "auto", see
                    `~greykite.algo.forecast.silverkite.forecast_silverkite_helper.get_silverkite_uncertainty_dict`
                    for the default value. If ``coverage`` is not None and ``uncertainty_dict`` is not provided,
                    then the "auto" setting is used.

                    If ``coverage`` is None and ``uncertainty_dict`` is None, then no intervals are returned.

            custom: `dict` [`str`, `any`] or None, optional
                Custom parameters that don't fit the categories above. Dictionary
                with the following optional keys:

                ``"fit_algorithm_dict"``: `dict` or a list of such values for grid search
                    How to fit the model. A dictionary with the following optional keys.

                        ``"fit_algorithm"``: `str`, optional, default "ridge"
                            The type of predictive model used in fitting.

                            See `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
                            for available options and their parameters.
                        ``"fit_algorithm_params"``: `dict` or None, optional, default None
                            Parameters passed to the requested fit_algorithm.
                            If None, uses the defaults in `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.

                ``"feature_sets_enabled"``: `dict` [`str`, `bool` or "auto" or None] or `bool` or "auto" or None; or a list of such values for grid search
                    Whether to include interaction terms and categorical variables to increase model flexibility.

                    If a `dict`, boolean values indicate whether include various sets of features in the model.
                    The following keys are recognized
                    (from `~greykite.algo.forecast.silverkite.constants.silverkite_column.SilverkiteColumn`):

                        ``"COLS_HOUR_OF_WEEK"``: `str`
                            Constant hour of week effect
                        ``"COLS_WEEKEND_SEAS"``: `str`
                            Daily seasonality interaction with is_weekend
                        ``"COLS_DAY_OF_WEEK_SEAS"``: `str`
                            Daily seasonality interaction with day of week
                        ``"COLS_TREND_DAILY_SEAS"``: `str`
                            Allow daily seasonality to change over time by is_weekend
                        ``"COLS_EVENT_SEAS"``: `str`
                            Allow sub-daily event effects
                        ``"COLS_EVENT_WEEKEND_SEAS"``: `str`
                            Allow sub-daily event effect to interact with is_weekend
                        ``"COLS_DAY_OF_WEEK"``: `str`
                            Constant day of week effect
                        ``"COLS_TREND_WEEKEND"``: `str`
                            Allow trend (growth, changepoints) to interact with is_weekend
                        ``"COLS_TREND_DAY_OF_WEEK"``: `str`
                            Allow trend to interact with day of week
                        ``"COLS_TREND_WEEKLY_SEAS"``: `str`
                            Allow weekly seasonality to change over time

                    The following dictionary values are recognized:

                        - True: include the feature set in the model
                        - False: do not include the feature set in the model
                        - None: do not include the feature set in the model
                        - "auto" or not provided: use the default setting based on data frequency and size

                    If not a `dict`:

                        - if a boolean, equivalent to a dictionary with all values set to the boolean.
                        - if None, equivalent to a dictionary with all values set to False.
                        - if "auto", equivalent to a dictionary with all values set to "auto".

                ``"max_daily_seas_interaction_order"``: `int` or None or a list of such values for grid search, default 5
                    Max fourier order to use for interactions with daily seasonality.
                    (COLS_EVENT_SEAS, COLS_EVENT_WEEKEND_SEAS, COLS_WEEKEND_SEAS, COLS_DAY_OF_WEEK_SEAS, COLS_TREND_DAILY_SEAS).

                    Model includes interactions terms specified by ``feature_sets_enabled``
                    up to the order limited by this value and the available order from ``seasonality``.
                ``"max_weekly_seas_interaction_order"`` : `int` or None or a list of such values for grid search, default 2
                    Max fourier order to use for interactions with weekly seasonality (COLS_TREND_WEEKLY_SEAS).

                    Model includes interactions terms specified by ``feature_sets_enabled``
                    up to the order limited by this value and the available order from ``seasonality``.
                ``"extra_pred_cols"``: `list` [`str`] or None or a list of such values for grid search, default None
                    Names of extra predictor columns to pass to ``forecast_silverkite``.
                    The standard interactions can be controlled via ``feature_sets_enabled`` parameter.
                    Accepts any valid patsy model formula term. Can be used to model complex interactions
                    of time features, events, seasonality, changepoints, regressors. Columns should be
                    generated by ``build_silverkite_features`` or included with input data.
                    These are added to any features already included by ``feature_sets_enabled`` and
                    terms specified by ``model``.
                ``"min_admissible_value"``: `float` or `double` or `int` or None, default None
                    The lowest admissible value for the forecasts and prediction
                    intervals. Any value below this will be mapped back to this value.
                    If None, there is no lower bound.
                ``"max_admissible_value"``: `float` or `double` or `int` or None, default None
                    The highest admissible value for the forecasts and prediction
                    intervals. Any value above this will be mapped back to this value.
                    If None, there is no upper bound.

            hyperparameter_override: `dict` [`str`, `any`] or None or `list` [`dict` [`str`, `any`] or None], optional
                After the above model components are used to create a hyperparameter grid, the result is
                updated by this dictionary, to create new keys or override existing ones.
                Allows for complete customization of the grid search.

                Keys should have format ``{named_step}__{parameter_name}`` for the named steps of the
                `sklearn.pipeline.Pipeline` returned by this function. See `sklearn.pipeline.Pipeline`.

                For example::

                    hyperparameter_override={
                        "estimator__silverkite": SimpleSilverkiteForecast(),
                        "estimator__silverkite_diagnostics": SilverkiteDiagnostics(),
                        "estimator__growth_term": "linear",
                        "input__response__null__impute_algorithm": "ts_interpolate",
                        "input__response__null__impute_params": {"orders": [7, 14]},
                        "input__regressors_numeric__normalize__normalize_algorithm": "RobustScaler",
                    }

                If a list of dictionaries, grid search will be done for each dictionary in the list.
                Each dictionary in the list override the defaults. This enables grid search
                over specific combinations of parameters to reduce the search space.

                * For example, the first dictionary could define combinations of parameters for a
                  "complex" model, and the second dictionary could define combinations of parameters
                  for a "simple" model, to prevent mixed combinations of simple and complex.
                * Or the first dictionary could grid search over fit algorithm, and the second dictionary
                  could use a single fit algorithm and grid search over seasonality.

                The result is passed as the ``param_distributions`` parameter
                to `sklearn.model_selection.RandomizedSearchCV`.

        model_template: `str`, `list`[`str`] or None, default None
            The simple silverkite template support single templates, multi templates or a list of single/multi templates.
            A valid single template must be either ``SILVERKITE`` or consists of

                {FREQ}_SEAS_{VAL}_GR_{VAL}_CP_{VAL}_HOL_{VAL}_FEASET_{VAL}_ALGO_{VAL}_AR_{VAL}

            For example, we have DAILY_SEAS_NM_GR_LINEAR_CP_LT_HOL_NONE_FEASET_ON_ALGO_RIDGE_AR_ON. The valid FREQ and VAL
            can be found at `~greykite.framework.templates.template_defaults`. The components stand for seasonality,
            growth, changepoints_dict, events, feature_sets_enabled, fit_algorithm and autoregression in
            `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`, which is used in
            `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`. Users are allowed to

                - Omit any number of component-value pairs, and the omitted will be filled with default values.
                - Switch the order of different component-value pairs.

            A valid multi template must belong to
            `~greykite.framework.templates.template_defaults.MULTI_TEMPLATES`
            or must be a list of single or multi template names.
    """
    DEFAULT_MODEL_TEMPLATE = "SILVERKITE"
    """The default model template. See `~greykite.framework.templates.model_templates.ModelTemplateEnum`.
    Uses a string to avoid circular imports.
    Overrides the value from `~greykite.framework.templates.forecast_config_defaults.ForecastConfigDefaults`.
    """
    def __init__(
            self,
            constants: SimpleSilverkiteTemplateConstants = SimpleSilverkiteTemplateConstants(),
            estimator: BaseForecastEstimator = SimpleSilverkiteEstimator()):
        super().__init__(estimator=estimator)
        self._constants = constants
        """Constants used by the template class. Includes the model templates and their default values."""

    @property
    def allow_model_template_list(self) -> bool:
        """SimpleSilverkiteTemplate allows config.model_template to be a list."""
        return True

    @property
    def allow_model_components_param_list(self) -> bool:
        """SilverkiteTemplate allows `config.model_components_param` to be a list."""
        return True

    @property
    def constants(self) -> SimpleSilverkiteTemplateConstants:
        """Constants used by the template class. Includes the model templates and their default values.
        """
        return self._constants

    @property
    def _silverkite_holiday(self) -> Type[SilverkiteHoliday]:
        """Holiday constants used by the estimator."""
        return self.estimator.silverkite._silverkite_holiday

    def get_regressor_cols(self):
        """Returns regressor column names from the model components.

        Implements the method in `~greykite.framework.templates.base_template.BaseTemplate`.

        Uses these attributes:

            model_components: :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`,
            `list` [`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`] or None, default None
                Configuration of model growth, seasonality, holidays, etc.
                See :func:`~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`
                for details.

        Returns
        -------
        regressor_cols : `list` [`str`] or None
            The names of regressor columns used in any hyperparameter set
            requested by ``model_components``.
            None if there are no regressors.
        """
        model_components = self.config.model_components_param
        if model_components is not None and isinstance(model_components, ModelComponentsParam) and model_components.regressors is not None:
            # ``regressor_cols`` is a list of strings to initialize
            # SimpleSilverkiteEstimator.regressor_cols, or a list of
            # such lists.
            regressor_cols = model_components.regressors.get("regressor_cols", [])
        elif model_components is not None and isinstance(model_components, list) and any([item.regressors is not None for item in model_components]):
            # Provides a list of ``model_components_param``.
            # ``regressor_cols`` is the union of each single ``regressor_cols``.
            regressor_cols = []
            for item in model_components:
                if item.regressors is not None:
                    regressor_cols += item.regressors.get("regressor_cols", [])
        else:
            regressor_cols = []
        return unique_in_list(
            array=regressor_cols,
            ignored_elements=(None,))

    def get_hyperparameter_grid(self):
        """Returns hyperparameter grid.

        Implements the method in `~greykite.framework.templates.base_template.BaseTemplate`.

        Converts model components, time properties, and model template into
        :class:`~greykite.sklearn.estimator.simple_silverkite_estimator.SimpleSilverkiteEstimator`
        hyperparameters.

        Uses these attributes:

            model_components: :class:`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`,
            `list` [`~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`] or None, default None
                Configuration of model growth, seasonality, events, etc.
                See :func:`~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`
                for details.

            time_properties: `dict` [`str`, `any`] or None, default None
                Time properties dictionary (likely produced by
                `~greykite.common.time_properties_forecast.get_forecast_time_properties`)
                with keys:

                ``"period"``: `int`
                    Period of each observation (i.e. minimum time between observations, in seconds).
                ``"simple_freq"``: `SimpleTimeFrequencyEnum`
                    ``SimpleTimeFrequencyEnum`` member corresponding to data frequency.
                ``"num_training_points"``: `int`
                    Number of observations for training.
                ``"num_training_days"``: `int`
                    Number of days for training.
                ``"start_year"``: `int`
                    Start year of the training period.
                ``"end_year"``: `int`
                    End year of the forecast period.
                ``"origin_for_time_vars"``: `float`
                    Continuous time representation of the first date in ``df``.

            model_template: `str`, default "SILVERKITE"
                The name of model template, must be one of the valid templates defined in
                `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`.

        Notes
        -----
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`
        handles the train/test splits according to ``EvaluationPeriodParam``,
        so ``estimator__train_test_thresh`` and ``estimator__training_fraction`` are always None.

        Similarly, ``estimator__origin_for_time_vars`` is set to None.

        Returns
        -------
        hyperparameter_grid : `dict` [`str`, `list` [`any`]] or `list` [ `dict` [`str`, `list` [`any`]] ]
            hyperparameter_grid for grid search in :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
            The output dictionary values are lists, combined in grid search.
        """
        if self.config.model_components_param is None or isinstance(self.config.model_components_param, ModelComponentsParam):
            model_components_list = [self.config.model_components_param]
        else:
            model_components_list = self.config.model_components_param
        model_components_list = self.__get_model_components_and_override_from_model_template(
            template=self.config.model_template,
            model_components=model_components_list)
        hyperparameter_grid_result = []
        for model_components in model_components_list:
            hyperparameter_grid = self.__get_hyperparameter_grid_from_model_components(model_components)
            # Adds time/training information to `hyperparameter_grid`.
            hyperparameter_grid = update_dictionaries(
                default_dict=hyperparameter_grid,
                overwrite_dicts={
                    "estimator__time_properties": [self.time_properties],
                    "estimator__origin_for_time_vars": [None],
                    "estimator__train_test_thresh": [None],
                    "estimator__training_fraction": [None]})
            # Update values with `hyperparameter_override`.
            hyperparameter_grid = update_dictionaries(
                default_dict=hyperparameter_grid,
                overwrite_dicts=model_components.hyperparameter_override)
            if isinstance(hyperparameter_grid, dict):
                hyperparameter_grid = [hyperparameter_grid]
            hyperparameter_grid_result += hyperparameter_grid
        # Ensures all items have the proper type for
        # `sklearn.model_selection.RandomizedSearchCV`.
        # List-type hyperparameters are specified below
        # with their accepted non-list type values.
        hyperparameter_grid_result = dictionaries_values_to_lists(
            hyperparameter_grid_result,
            hyperparameters_list_type={
                "estimator__holidays_to_model_separately": [None, "auto", self._silverkite_holiday.ALL_HOLIDAYS_IN_COUNTRIES],
                "estimator__holiday_lookup_countries": [None, "auto"],
                "estimator__regressor_cols": [None],
                "estimator__extra_pred_cols": [None]}
        )
        hyperparameter_grid_result = unique_dict_in_list(hyperparameter_grid_result)
        if len(hyperparameter_grid_result) == 1:
            hyperparameter_grid_result = hyperparameter_grid_result[0]
        return hyperparameter_grid_result

    # Additional helper functions for `get_hyperparameter_grid` are below.
    # Their behavior is customized by `self.constants`.

    # Uses `self.constants`.
    def __template_name_from_dataclass(self, name_dataclass=None):
        """Decodes the string format single template name from the ``name_dataclass`` dataclass.

        Parameters
        ----------
        name_dataclass : `SimpleSilverkiteTemplateName`, default None.
            The dataclass that contains the simple silverkite template string-type name values.
            If None, the default values will be used.

        Returns
        -------
        full_name : `str`
            The processed template name. For example, the input above will become
            `DAILY_SEAS_LT_GR_LINEAR_CP_LT_HOL_NONE_FEASET_OFF_ALGO_LINEAR_AR_ON`.
        """
        if name_dataclass is None:
            name_dataclass = self.constants.SimpleSilverkiteTemplateOptions()
        return f"{name_dataclass.freq.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.SEAS.name}_{name_dataclass.seas.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.GR.name}_{name_dataclass.gr.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.CP.name}_{name_dataclass.cp.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.HOL.name}_{name_dataclass.hol.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.FEASET.name}_{name_dataclass.feaset.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.ALGO.name}_{name_dataclass.algo.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.AR.name}_{name_dataclass.ar.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.DSI.name}_{name_dataclass.dsi.name}_" \
               f"{self.constants.SILVERKITE_COMPONENT_KEYWORDS.WSI.name}_{name_dataclass.wsi.name}"

    # Uses `self.constants`.
    def __decode_single_template(self, template):
        """Given a single simple silverkite template's name, preprocess the name order and fill in default values.

        The naming of a simple silverkite template is either ``SILVERKITE`` or consists of

            {FREQ}_SEAS_{VAL}_GR_{VAL}_CP_{VAL}_HOL_{VAL}_FEASET_{VAL}_ALGO_{VAL}_AR_{VAL}

        For example, we have DAILY_SEAS_NM_GR_LINEAR_CP_LT_HOL_NONE_FEASET_ON_ALGO_RIDGE_AR_ON. The valid FREQ and VAL
        can be found at `~greykite.framework.templates.template_defaults`. The components stand for seasonality,
        growth, changepoints_dict, events, feature_sets_enabled, fit_algorithm and autoregression in
        `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`, which is used in
        `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`. Users are allowed to

            - Omit any number of component-value pairs, and the omitted will be filled with default values.
            - Switch the order of different component-value pairs.

        Parameters
        ----------
        template : `str` or `SimpleSilverkiteTemplateName`
            A template name. For example, `DAILY_CP_LT_HOL_NONE_AR_ON`.
            You could also define the name throught the `SimpleSilverkiteTemplateName` class.

        Returns
        -------
        full_name : `str`
            The processed template name. For example, the input above will become
            `DAILY_SEAS_LT_GR_LINEAR_CP_LT_HOL_NONE_FEASET_OFF_ALGO_LINEAR_AR_ON`.
        """
        if template == "SILVERKITE":
            return "SILVERKITE"
        if template == "SILVERKITE_EMPTY":
            return self.constants.SILVERKITE_EMPTY
        if isinstance(template, str):
            names = template.split("_")
            freq, components = names[0], names[1:]
            if freq not in self.constants.VALID_FREQ:
                raise ValueError(f"Template not recognized. Please make sure the frequency belongs to {self.constants.VALID_FREQ}.")
            new_name = [freq]
            # No keyword in the template name, try to decode it with the default order.
            # If `components` is empty, then all values are filled with default.
            if not any([component in self.constants.SILVERKITE_COMPONENT_KEYWORDS.__members__ for component in components]) and components:
                raise ValueError("No keyword found in the template name. Please make sure your template name "
                                 "consists of keyword + value pairs.")
            # At least one keyword in the template name. Try to fill the others with default values.
            # In this case the order of the components does not matter, so we iterate the OrderedDict.
            used_components = []
            default_components = []
            # Fills component names and values in the default order.
            # Uses the input value if exists.
            for component in self.constants.SILVERKITE_COMPONENT_KEYWORDS.__members__:
                if component == "FREQ":
                    continue
                if component in components:
                    i = components.index(component)
                    try:
                        value = components[i + 1]
                        if value not in self.constants.SILVERKITE_COMPONENT_KEYWORDS[component].value.__members__:
                            raise ValueError(f"{value} is not recognized for {component}. "
                                             f"The valid values are {list(self.constants.SILVERKITE_COMPONENT_KEYWORDS[component].value.__members__)}.")
                    except IndexError:
                        raise ValueError("Template name is not valid. It must be either keyword + value or purely values.")
                    used_components.append(i)
                    used_components.append(i + 1)
                else:
                    value = self.constants.SILVERKITE_COMPONENT_KEYWORDS[component].value.DEFAULT.name
                    default_components.append(component)
                new_name.append(component)
                new_name.append(value)
            unused_components = [component for i, component in enumerate(components) if i not in used_components]
            if unused_components:
                warnings.warn("The following words are not used because they are neither a component keyword"
                              f" nor following a component keyword, or because you have duplicate keywords. {unused_components}")
            default_components = [item for item in default_components if item not in ["DSI", "WSI"]]  # These two keywords are not exposed to users.
            if default_components:
                warnings.warn("The following component keywords are not found in the template name, "
                              "thus the default values will be used. For default values, please check the doc. "
                              f"{default_components}")
            return "_".join(new_name)
        elif isinstance(template, self.constants.SimpleSilverkiteTemplateOptions):
            return self.__template_name_from_dataclass(template)
        else:
            raise ValueError(f"The template type {type(template)} is not recognized. It must be str or `SimpleSilverkiteTemplateStringName`.")

    # Uses `self.constants`, and called by `forecaster.py`
    def check_template_type(self, template):
        """Checks the template name is valid and whether it is single or multi template.
        Raises an error if the template is not recognized.

        A valid single template must be either ``SILVERKITE`` or consists of

            {FREQ}_SEAS_{VAL}_GR_{VAL}_CP_{VAL}_HOL_{VAL}_FEASET_{VAL}_ALGO_{VAL}_AR_{VAL}

        For example, we have DAILY_SEAS_NM_GR_LINEAR_CP_LT_HOL_NONE_FEASET_ON_ALGO_RIDGE_AR_ON. The valid FREQ and VAL
        can be found at `~greykite.framework.templates.template_defaults`. The components stand for seasonality,
        growth, changepoints_dict, events, feature_sets_enabled, fit_algorithm and autoregression in
        `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`, which is used in
        `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`. Users are allowed to

            - Omit any number of component-value pairs, and the omitted will be filled with default values.
            - Switch the order of different component-value pairs.

        A valid multi template must belong to
        `~greykite.framework.templates.template_defaults.MULTI_TEMPLATES`
        or must be a list of single or multi template names.

        Parameters
        ----------
        template : `str`, `SimpleSilverkiteTemplateName` or `list`[`str`, `SimpleSilverkiteTemplateName`]
            The ``model_template`` parameter fed into
            `~greykite.framework.templates.autogen.forecast_config.ForecastConfig`.
            for simple silverkite templates.

        Returns
        -------
        template_type : `str`
            "single" or "multi".
        """
        if (template in ["SILVERKITE", "SILVERKITE_EMPTY"]
                or isinstance(template, self.constants.SimpleSilverkiteTemplateOptions)
                or template.split("_")[0] in self.constants.VALID_FREQ):
            return "single"
        if template in self.constants.MULTI_TEMPLATES:
            return "multi"
        raise ValueError(f"The template name {template} is not recognized. It must be 'SILVERKITE', 'SILVERKITE_EMPTY', "
                         f"a `SimpleSilverkiteTemplateOptions` data class, of the type"
                         " '{FREQ}_SEAS_{VAL}_GR_{VAL}_CP_{VAL}_HOL_{VAL}_FEASET_{VAL}_ALGO_{VAL}_AR_{VAL}' or"
                         f" belong to {list(self.constants.MULTI_TEMPLATES.keys())}.")

    # Depends on a function that uses `self.constants`.
    def __get_name_string_from_model_template(self, template):
        """Gets the name string(s) from model template.

        The template could be a name string, a `SimpleSilverkiteTemplateOptions` dataclass,
        or a list of such strings and/or dataclasses.
        If a list is given, a list of `ModelComponentsParam` is returned.

        The naming of a simple silverkite template is either ``SILVERKITE`` or consists of

            {FREQ}_SEAS_{VAL}_GR_{VAL}_CP_{VAL}_HOL_{VAL}_FEASET_{VAL}_ALGO_{VAL}_AR_{VAL}

        For example, we have DAILY_SEAS_NM_GR_LINEAR_CP_LT_HOL_NONE_FEASET_ON_ALGO_RIDGE_AR_ON. The valid FREQ and VAL
        can be found at `~greykite.framework.templates.template_defaults`. The components stand for seasonality,
        growth, changepoints_dict, events, feature_sets_enabled, fit_algorithm and autoregression in
        `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`, which is used in
        `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`. Users are allowed to

            - Omit any number of component-value pairs, and the omitted will be filled with default values.
            - Switch the order of different component-value pairs.

        Parameters
        ----------
        template : `str`, `SimpleSilverkiteTemplateOptions` or `list` [`str`, `SimpleSilverkiteTemplateOptions`]
            The ``model_template`` in `ForecastConfig`, could be a name string, a `SimpleSilverkiteTemplateOptions` dataclass,
            or a list of such strings and/or dataclasses.

        Returns
        -------
        name_string: `list` [`str`]
            The list of name string(s) that corresponds to ``template``.
        """
        def get_name_string_from_single_or_multi_template(model_template):
            """Gets the name string when model template type is single or multi."""
            template_type = self.check_template_type(model_template)
            if template_type == "single":
                return [self.__decode_single_template(model_template)]
            if template_type == "multi":
                return [self.__decode_single_template(single_template)
                        for single_template in self.constants.MULTI_TEMPLATES[model_template]]

        if isinstance(template, list):
            name_strings = []
            for item in template:
                if isinstance(item, list):
                    raise ValueError(f"{item} is not recognized as a valid single or multi template name.")
                current_item_name_string = get_name_string_from_single_or_multi_template(item)
                name_strings += current_item_name_string
            return name_strings
        else:
            name_strings = get_name_string_from_single_or_multi_template(template)
            return name_strings

    # Uses `self.constants`.
    def __get_single_model_components_param_from_template(self, template):
        """Gets the `ModelComponentsParam` class from a single simple silverkite template name.

        Parameters
        ----------
        template : `str`
            The full template name output by
            `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate.__decode_single_template`.

        Returns
        -------
        model_components_param : `ModelComponentsParam`
            The corresponding `ModelComponentsParam`, see `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam`.
        """
        if template == "SILVERKITE":
            return self.constants.SILVERKITE
        names = template.split("_")
        freq, components = names[0], names[1:]
        # The value of a keyword is supposed to follow it.
        # For each keyword, we extract the next substring (index + 1) as its value.
        model_components_param = ModelComponentsParam(
            seasonality=self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["SEAS"][freq][components[components.index("SEAS")+1]],
            growth=self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["GR"][components[components.index("GR")+1]],
            changepoints={
                "changepoints_dict": self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["CP"][freq][components[components.index("CP")+1]],
                "seasonality_changepoints_dict": None
            },
            events=self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["HOL"][components[components.index("HOL")+1]],
            custom={
                "feature_sets_enabled": self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["FEASET"][components[components.index("FEASET")+1]],
                "fit_algorithm_dict": self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["ALGO"][components[components.index("ALGO")+1]],
                "max_daily_seas_interaction_order": self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["DSI"][freq][components[components.index("DSI")+1]],
                "max_weekly_seas_interaction_order": self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["WSI"][freq][components[components.index("WSI")+1]],
                "extra_pred_cols": [],
                "min_admissible_value": None,
                "max_admissible_value": None
            },
            autoregression=self.constants.COMMON_MODELCOMPONENTPARAM_PARAMETERS["AR"][components[components.index("AR")+1]],
            regressors={
                "regressor_cols": []
            },
            uncertainty={
                "uncertainty_dict": None
            })
        return model_components_param

    # Public interface, useful for debugging.
    def get_model_components_from_model_template(self, template):
        """Gets the `ModelComponentsParam` class from model template.

        The template could be a name string, a `SimpleSilverkiteTemplateOptions` dataclass,
        or a list of such strings and/or dataclasses.
        If a list is given, a list of `ModelComponentsParam` is returned.
        If a single element is given, a list of length 1 is returned.

        Parameters
        ----------
        template : `str`, `SimpleSilverkiteTemplateOptions` or `list` [`str`, `SimpleSilverkiteTemplateOptions`]
            The ``model_template`` in `ForecastConfig`, could be a name string, a `SimpleSilverkiteTemplateOptions` dataclass,
            or a list of such strings and/or dataclasses.

        Returns
        -------
        model_components_param: `list` [`ModelComponentsParam`]
            The list of `ModelComponentsParam` class(es) that correspond to ``template``.
        """
        name_strings = self.__get_name_string_from_model_template(template)
        model_components = [self.__get_single_model_components_param_from_template(single_template) for single_template in name_strings]
        return model_components

    # Depends on a function that uses `self.constants`.
    def __override_model_components(
            self,
            default_model_components,
            model_components=None):
        """Uses the values in ``model_components`` to update the corresponding values in ``default_model_components``.

        If a value is not set in model_components, i.e., the value is None, then the default value is not updated.

        Parameters
        ----------
        default_model_components : `ModelComponentsParam`
            The default model components to be updated.
        model_components : `ModelComponentsParam`, default None
            The new model components used to update ``default_model_components``.

        Returns
        -------
        updated_model_components : `ModelComponentsParam`
            The updated model components.
        """
        default_model_components = dataclasses.replace(default_model_components)  # Returns a copy.
        if default_model_components.hyperparameter_override is None:
            default_model_components.hyperparameter_override = {}
        if model_components is None:
            return default_model_components
        components = ["seasonality", "growth", "changepoints", "events", "custom", "autoregression", "regressors", "uncertainty", "hyperparameter_override"]
        for component in components:
            allow_unknown_keys = (component == "hyperparameter_override")  # Allows unknown keys for hyperparameter_override.
            updated_value = update_dictionaries(
                default_dict=getattr(default_model_components, component, {}),
                overwrite_dicts=getattr(model_components, component),
                allow_unknown_keys=allow_unknown_keys)
            setattr(default_model_components, component, updated_value)
        return default_model_components

    # Depends on a function that uses `self.constants`.
    def __get_model_components_and_override_from_model_template(
            self,
            template,
            model_components=None):
        """Gets the `ModelComponentsParam` class from model template, and overridden by ``model_components``.

        The template could be a name string, a `SimpleSilverkiteTemplateOptions` dataclass,
        or a list of such strings and/or dataclasses.
        If a list is given, a list of `ModelComponentsParam` is returned.
        If a single element is give, a list of length 1 is returned.

        The full cross product of all templates overridden by all provided model components is returned.

        Parameters
        ----------
        template : `str`, `SimpleSilverkiteTemplateOptions` or `list` [`str`, `SimpleSilverkiteTemplateOptions`]
            The ``model_template`` in `ForecastConfig`, could be a name string, a `SimpleSilverkiteTemplateOptions` dataclass,
            or a list of such strings and/or dataclasses.

        model_components : `ModelComponentsParam` or `list` [`ModelComponentsParam`], default None.
            The `ModelComponentsParam` in `ForecastConfig`, used to override the default values provided by ``template``.

        Returns
        -------
        model_components_param: `list` [`ModelComponentsParam`]
            The list of `ModelComponentsParam` class(es) that correspond to ``template``, overridden by ``model_components``.
        """
        if model_components is None:
            model_components = []
        if isinstance(model_components, ModelComponentsParam):
            model_components = [model_components]
        default_model_components = self.get_model_components_from_model_template(template)
        new_model_components = [
            self.__override_model_components(single_default_model_components, single_model_components)
            for single_model_components in model_components for single_default_model_components in default_model_components]
        return new_model_components

    @staticmethod
    def __get_hyperparameter_grid_from_model_components(model_components):
        """Gets the hyperparameter grid from the `ModelComponentsParam` class.

        Parameters
        ----------
        model_components : `class`
            The `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam` class.
            Configuration of model growth, seasonality, events, etc.
            See :func:`~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`
            for details.

        Returns
        -------
        hyperparameter_grid : `dict`
            hyperparameter_grid for hyperparameter_override in
             `~greykite.framework.templates.autogen.forecast_config.ModelComponentsParam.hyperparameter_override`.
        """
        hyperparameter_grid = {
            "estimator__yearly_seasonality": model_components.seasonality["yearly_seasonality"],
            "estimator__quarterly_seasonality": model_components.seasonality["quarterly_seasonality"],
            "estimator__monthly_seasonality": model_components.seasonality["monthly_seasonality"],
            "estimator__weekly_seasonality": model_components.seasonality["weekly_seasonality"],
            "estimator__daily_seasonality": model_components.seasonality["daily_seasonality"],
            "estimator__growth_term": model_components.growth["growth_term"],
            "estimator__changepoints_dict": model_components.changepoints["changepoints_dict"],
            "estimator__seasonality_changepoints_dict": model_components.changepoints["seasonality_changepoints_dict"],
            "estimator__holidays_to_model_separately": model_components.events["holidays_to_model_separately"],
            "estimator__holiday_lookup_countries": model_components.events["holiday_lookup_countries"],
            "estimator__holiday_pre_num_days": model_components.events["holiday_pre_num_days"],
            "estimator__holiday_post_num_days": model_components.events["holiday_post_num_days"],
            "estimator__holiday_pre_post_num_dict": model_components.events["holiday_pre_post_num_dict"],
            "estimator__daily_event_df_dict": model_components.events["daily_event_df_dict"],
            "estimator__feature_sets_enabled": model_components.custom["feature_sets_enabled"],
            "estimator__fit_algorithm_dict": model_components.custom["fit_algorithm_dict"],
            "estimator__max_daily_seas_interaction_order": model_components.custom["max_daily_seas_interaction_order"],
            "estimator__max_weekly_seas_interaction_order": model_components.custom["max_weekly_seas_interaction_order"],
            "estimator__extra_pred_cols": model_components.custom["extra_pred_cols"],
            "estimator__min_admissible_value": model_components.custom["min_admissible_value"],
            "estimator__max_admissible_value": model_components.custom["max_admissible_value"],
            "estimator__autoreg_dict": model_components.autoregression["autoreg_dict"],
            "estimator__regressor_cols": model_components.regressors["regressor_cols"],
            "estimator__uncertainty_dict": model_components.uncertainty["uncertainty_dict"]
        }
        return hyperparameter_grid



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# original author: Rachit Arora

from dataclasses import dataclass
from enum import Enum
from typing import Type

from greykite.common.enums import TimeEnum


@dataclass
class SilverkiteSeasonality:
    """Contains information to create `fs_components_df` parameter for `forecast_silverkite` for modeling seasonality."""

    name: str
    """Name of the timeseries feature (e.g. tod, tow etc.)"""

    period: float
    """Period of the fourier series"""

    order: int
    """Recommended order of the fourier series. Add interactions to increase flexibility"""

    seas_names: str
    """Additional label for the seasonality term"""

    default_min_days: int
    """Recommended minimum input data size before adding the seasonality"""


class SilverkiteSeasonalityEnum(Enum):
    """Defines default seasonalities for Silverkite estimator.
    Names should match those in SeasonalityEnum.
    The default order for various seasonalities is stored in this enum.
    """

    DAILY_SEASONALITY: SilverkiteSeasonality = SilverkiteSeasonality(
        name="tod",
        period=24.0,
        order=12,
        seas_names="daily",
        default_min_days=2)
    """``tod`` is 0-24 time of day (tod granularity based on input data, up to second level).
    Requires at least two full cycles to add the seasonal term (``default_min_days=2``).
    """

    WEEKLY_SEASONALITY: SilverkiteSeasonality = SilverkiteSeasonality(
        name="tow",
        period=7.0,
        order=4,
        seas_names="weekly",
        default_min_days=TimeEnum.ONE_WEEK_IN_DAYS.value * 2)
    """``tow`` is 0-7 time of week (tow granularity based on input data, up to second level).
    ``order=4`` for full flexibility to model daily input.
    """

    MONTHLY_SEASONALITY: SilverkiteSeasonality = SilverkiteSeasonality(
        name="tom",
        period=1.0,
        order=2,
        seas_names="monthly",
        default_min_days=TimeEnum.ONE_MONTH_IN_DAYS.value * 2)
    """``tom`` is 0-1 time of month (tom granularity based on input data, up to daily level)."""

    QUARTERLY_SEASONALITY: SilverkiteSeasonality = SilverkiteSeasonality(
        name="toq",
        period=1.0,
        order=5,
        seas_names="quarterly",
        default_min_days=TimeEnum.ONE_QUARTER_IN_DAYS.value * 2)
    """``toq`` (continuous time of quarter) with natural period.
    Each day is mapped to a value in [0.0, 1.0) based on its position in the calendar quarter:
    (Jan1-Mar31, Apr1-Jun30, Jul1-Sep30, Oct1-Dec31). The start of each quarter is 0.0.
    """

    YEARLY_SEASONALITY: SilverkiteSeasonality = SilverkiteSeasonality(
        name="ct1",
        period=1.0,
        order=15,
        seas_names="yearly",
        default_min_days=round(TimeEnum.ONE_YEAR_IN_DAYS.value * 1.5))
    """``ct1`` (continuous year) with natural period."""


class SilverkiteSeasonalityEnumMixin:
    """Provides a mixin for the SilverkiteSeasonalityEnum constants"""

    def get_silverkite_seasonality_enum(self) -> Type[SilverkiteSeasonalityEnum]:
        """Return the SilverkiteSeasonalityEnum constants"""
        return SilverkiteSeasonalityEnum



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""All forecast estimators must inherit from this class."""

from abc import ABC
from abc import abstractmethod

from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin
from sklearn.metrics import mean_squared_error

from greykite.common import constants as cst
from greykite.common.evaluation import r2_null_model_score
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.sklearn.estimator.null_model import DummyEstimator


class BaseForecastEstimator(BaseEstimator, RegressorMixin, ABC):
    """A base class for forecast models. Fits a timeseries and predicts future values

    Parameters
    ----------
    score_func : callable, optional, default=mean_squared_error
        Function to calculate model R2_null_model_score score,
        with signature (actual, predicted).
        `actual`, `predicted` are array-like with the same shape.
        Smaller values are better.

    coverage : float, optional, default=0.95
        intended coverage of the prediction bands (0.0 to 1.0)
        If None, the upper/lower predictions are not returned by `predict`

        Every subclass must use `coverage` to set prediction band width. This ensures a common
        BaseForecastEstimator interface for parameters used during fitting and forecast evaluation

    null_model_params : dict with arguments passed to DummyRegressor, optional, default=None
        Dictionary keys must be in ("strategy", "constant", "quantile")
        Defines null model. model score is R2_null_model_score of model error relative to null model, evaluated by score_func
        If None, model score is score_func of the model itself

    Attributes
    ----------
    null_model : DummyEstimator
        null model used to measure model score
    time_col_ : str
        Name of input data time column
    value_col_ : str
        Name of input data value column
    last_predicted_X_ : `pandas.DataFrame` or None
        The ``X`` last passed to ``self.predict()``.
        Used to speed up predictions if the same ``X`` is passed repeatedly.
        Resets to None when ``self.fit()`` is called.
    cached_predictions_ : `pandas.DataFrame` or None
        The return value of the last call to ``self.predict()``.
        Used to speed up predictions if the same ``X`` is passed repeatedly.
        Resets to None when ``self.fit()`` is called.
    """
    @abstractmethod
    def __init__(self, score_func=mean_squared_error, coverage=0.95, null_model_params=None):
        """Initializes attributes common to every BaseForecastEstimator

        Subclasses must also have these parameters. Every subclass must call:

            super().__init__(score_func=score_func, coverage=coverage, null_model_params=null_model_params)

        """
        self.score_func = score_func
        self.coverage = coverage
        self.null_model_params = null_model_params

        # initializes attributes defined in fit
        self.null_model = None
        self.time_col_ = None
        self.value_col_ = None

        # initializes attributes defined in predict
        self.last_predicted_X_ = None    # the most recent X passed to self.predict()
        self.cached_predictions_ = None  # the most recent return value of self.predict()

    @abstractmethod
    def fit(self, X, y=None, time_col=cst.TIME_COL, value_col=cst.VALUE_COL, **fit_params):
        """Fits a model to training data
        Also fits the null model, if specified, for use in evaluating the `score` function

        Every subclass must call this::
            super().fit(X, y=y, time_col=time_col, value_col=value_col, **fit_params)

        Parameters
        ----------
        X : `pandas.DataFrame`
            Input timeseries, with timestamp column,
            value column, and any additional regressors.
            The value column is the response, included in
            ``X`` to allow transformation by `sklearn.pipeline`.
        y : ignored
            The original timeseries values, ignored.
            (The y for fitting is included in X.)
        time_col : `str`
            Time column name in X.
        value_col : `str`
            Value column name in X.
        fit_params : `dict`
            Additional parameters supported by subclass `fit` or null model.
        """
        self.time_col_ = time_col  # to be used in `predict` to select proper column
        self.value_col_ = value_col
        # Null model must be initialized here, otherwise scikit-learn
        # grid search will not be able to set the parameters.
        # See https://scikit-learn.org/stable/developers/develop.html#instantiation.
        if self.null_model_params is not None:
            # Adds score function to null model parameters, and initializes null model
            self.null_model_params["score_func"] = self.score_func
            self.null_model = DummyEstimator(**self.null_model_params)
            # Passes `sample_weight` rather than `**fit_params` to avoid unexpected keyword argument from the main
            #   estimator's parameters
            sample_weight = fit_params.get("sample_weight")
            self.null_model.fit(X, y=y, time_col=time_col, value_col=value_col, sample_weight=sample_weight)

        # Clears the cached result, because it is no longer valid for the updated model
        self.last_predicted_X_ = None
        self.cached_predictions_ = None

    @abstractmethod
    def predict(self, X, y=None):
        """Creates forecast for dates specified in X

        To enable caching, every subclass must call this at the beginning
        of its ``.predict()``. Before returning the result, the subclass
        ``.predict()`` must set ``self.cached_predictions_`` to the return value.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Input timeseries with timestamp column and any additional regressors.
            Timestamps are the dates for prediction.
            Value column, if provided in X, is ignored.
        y : ignored

        Returns
        -------
        predictions : `pandas.DataFrame`
            Forecasted values for the dates in X. Columns:

                - TIME_COL dates
                - PREDICTED_COL predictions
                - PREDICTED_LOWER_COL lower bound of predictions, optional
                - PREDICTED_UPPER_COL upper bound of predictions, optional
                - [other columns], optional

            ``PREDICTED_LOWER_COL`` and ``PREDICTED_UPPER_COL`` are present
            if ``self.coverage`` is not None.
        """
        if self.cached_predictions_ is not None and X.equals(self.last_predicted_X_):
            log_message("Returning cached predictions.", LoggingLevelEnum.DEBUG)
            return self.cached_predictions_
        else:
            # Updates `last_predicted_X` to the new value.
            # To enable caching, the subclass must set
            # `self.cached_predictions` to the returned result.
            self.last_predicted_X_ = X
            return None

    def summary(self):
        """Creates human readable string of how the model works, including relevant diagnostics
        These details cannot be extracted from the forecast alone
        Prints model configuration. Extend this in child class to print the trained model parameters.

        Log message is printed to the cst.LOGGER_NAME logger.
        """
        log_message(self, LoggingLevelEnum.DEBUG)  # print model input parameters

    def score(self, X, y, sample_weight=None):
        """Default scorer for the estimator (Used in GridSearchCV/RandomizedSearchCV if scoring=None)

        Notes
        -----
        If null_model_params is not None, returns R2_null_model_score of model error
        relative to null model, evaluated by score_func.

        If null_model_params is None, returns score_func of the model itself.

        By default, grid search (with no `scoring` parameter) optimizes improvement of ``score_func``
        against null model.

        To optimize a different score function, pass `scoring` to GridSearchCV/RandomizedSearchCV.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Input timeseries with timestamp column and any additional regressors.
            Value column, if provided in X, is ignored
        y : `pandas.Series` or  `numpy.array`
            Actual value, used to compute error
        sample_weight : `pandas.Series` or  `numpy.array`
            ignored

        Returns
        -------
        score : `float` or None
            Comparison of predictions against null predictions, according to specified score function
        """
        y_pred = self.predict(X)[cst.PREDICTED_COL]
        if self.null_model is not None:
            y_pred_null = self.null_model.predict(X)[cst.PREDICTED_COL]
            score = r2_null_model_score(
                y,
                y_pred,
                y_pred_null=y_pred_null,
                loss_func=self.score_func)
        else:
            score = self.score_func(y, y_pred)
        return score



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.exceptions import NotFittedError

from greykite.common.constants import TIME_COL
from greykite.common.features.timeseries_features import build_time_features_df
from greykite.common.features.timeseries_features import convert_date_to_continuous_time


class BuildTimeseriesFeaturesTransformer(BaseEstimator, TransformerMixin):
    """Calculates time series features (e.g. year, month, hour etc.) of the input time series

    Parameters
    ----------
    time_col : string, default=TIME_COL
        issues warning if fraction of nulls is above this value

    Attributes
    ----------
    origin_for_time_vars : float (e.g. 2019.2)
        sets default origin so that "ct1" feature from `build_time_features_df` starts at 0 on start date of fitted data
    """
    def __init__(self, time_col: str = TIME_COL):
        self.time_col = time_col
        self.origin_for_time_vars = None

    def fit(self, X, y=None):
        """Sets the time origin for input time series"""
        assert isinstance(X, pd.DataFrame)
        dt = X[self.time_col]
        self.origin_for_time_vars = convert_date_to_continuous_time(dt[0])
        return self

    def transform(self, X):
        """ Calculates time series features of the input time series

        Parameters
        ----------
        X : pd.DataFrame

        Returns
        -------
        A copy of the data frame with original time points and calculated features
        """
        if self.origin_for_time_vars is None:
            raise NotFittedError(
                "This instance is not fitted yet. Call 'fit' with appropriate arguments "
                "before calling 'transform'.")
        assert isinstance(X, pd.DataFrame)
        dt = X[self.time_col]
        features_ts = build_time_features_df(dt, conti_year_origin=self.origin_for_time_vars)
        output = pd.concat([dt, features_ts], axis=1)
        return output



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import MaxAbsScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import Normalizer
from sklearn.preprocessing import PowerTransformer
from sklearn.preprocessing import QuantileTransformer
from sklearn.preprocessing import RobustScaler
from sklearn.preprocessing import StandardScaler


NORMALIZE_ALGORITHMS = {
    "MinMaxScaler": MinMaxScaler,
    "MaxAbsScaler": MaxAbsScaler,
    "StandardScaler": StandardScaler,
    "RobustScaler": RobustScaler,
    "Normalizer": Normalizer,
    "QuantileTransformer": QuantileTransformer,
    "PowerTransformer": PowerTransformer,
}


class NormalizeTransformer(BaseEstimator, TransformerMixin):
    """Normalizes time series data.

    Parameters
    ----------
    normalize_algorithm : `str` or None, default None
        Which algorithm to use. Valid options are:

            - "MinMaxScaler" : `sklearn.preprocessing.MinMaxScaler`,
            - "MaxAbsScaler" : `sklearn.preprocessing.MaxAbsScaler`,
            - "StandardScaler" : `sklearn.preprocessing.StandardScaler`,
            - "RobustScaler" : `sklearn.preprocessing.RobustScaler`,
            - "Normalizer" : `sklearn.preprocessing.Normalizer`,
            - "QuantileTransformer" : `sklearn.preprocessing.QuantileTransformer`,
            - "PowerTransformer" : `sklearn.preprocessing.PowerTransformer`,

        If None, this transformer is a no-op. No normalization is done.

    normalize_params : `dict` or None, default None
        Params to initialize the normalization scaler/transformer.

    Attributes
    ----------
    scaler : `class`
        sklearn class used for normalization
    _is_fitted : `bool`
        Whether the transformer is fitted.
    """

    def __init__(
            self,
            normalize_algorithm=None,
            normalize_params=None):
        # sets params without modification to ensure get_params() works in grid search
        self.normalize_algorithm = normalize_algorithm
        self.normalize_params = normalize_params

        self.scaler = None
        self._is_fitted = None

    def fit(self, X, y=None):
        """Fits the normalization transform.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Training input data. e.g. each column is a timeseries.
            Columns are expected to be numeric.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        assert isinstance(X, pd.DataFrame)
        self._is_fitted = True
        if self.normalize_algorithm is not None:
            if self.normalize_algorithm not in NORMALIZE_ALGORITHMS.keys():
                raise ValueError(
                    f"`normalize_algorithm` '{self.normalize_algorithm}' is not recognized. "
                    f"Must be one of {NORMALIZE_ALGORITHMS.keys()}")
            if self.normalize_params is None:
                self.normalize_params = {}
            self.scaler = NORMALIZE_ALGORITHMS[self.normalize_algorithm](**self.normalize_params)
            self.scaler.fit(X=X)
        return self

    def transform(self, X):
        """Normalizes data using the specified scaling method.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Data to transform. e.g. each column is a timeseries.
            Columns are expected to be numeric.

        Returns
        -------
        X_normalized : `pandas.DataFrame`
            A normalized copy of the data frame.
        """
        if self._is_fitted is None:
            raise NotFittedError(
                "This instance is not fitted yet. Call 'fit' with appropriate arguments "
                "before calling 'transform'.")
        assert isinstance(X, pd.DataFrame)
        if self.scaler:
            transformed = self.scaler.transform(X)
            X_normalized = pd.DataFrame(transformed, index=X.index, columns=X.columns)
        else:
            X_normalized = X.copy()
        return X_normalized
